# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

import logging
from scullery import messagebus, workers

__doc__ = ""


import weakref
import threading
import functools
import os
import time
import threading
import traceback
import sys
import weakref
from .jsonrpyc import RPC
from subprocess import PIPE, STDOUT, Popen


def close_fds(p: Popen):
    try:
        p.stdin.close()
    except Exception:
        pass
    try:
        p.stdout.close()
    except Exception:
        pass
    try:
        p.stderr.close()
    except Exception:
        pass


@functools.cache
def which(program):
    "Check if a program is installed like you would do with UNIX's which command."

    # Because in windows, the actual executable name has .exe while the command name does not.
    if sys.platform == "win32" and not program.endswith(".exe"):
        program += ".exe"

    # Find out if path represents a file that the current user can execute.
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    # If the input was a direct path to an executable, return it
    if fpath:
        if is_exe(program):
            return program

    # Else search the path for the file.
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    # If we got this far in execution, we assume the file is not there and return None
    return None


# Util is not used anywhere else

# This is an acceptable dependamcy, it will be part of libkaithem if such a thing exists


# These events have to happen in a consistent order, the same order that the actual JACK callbacks happen.
# We can't do them within the jack callback becauuse that could do a deadlock.

jackEventHandlingQueue = []
jackEventHandlingQueueLock = threading.Lock()

portInfoByID = weakref.WeakValueDictionary()


class PortInfo:
    def __init__(self, name, isInput, sname, isAudio, aliases=None):
        self.isOutput = self.is_output = not isInput
        self.isInput = self.is_input = isInput
        self.isAudio = self.is_audio = isAudio
        self.name = name
        self.shortname = sname
        self.clientName = name[: -len(":" + sname)]
        portInfoByID[id(self)] = self
        self.aliases = aliases or []

    def toDict(self):
        return {
            "name": self.name,
            "isInput": self.is_input,
            "sname": self.shortname,
            "isAudio": self.isAudio,
            "aliases": self.aliases,
        }


class JackClientProxy:
    def __getattr__(self, attr):
        if self.ended or not self.worker.poll() is None:
            raise RuntimeError("This process is already dead")
        check_exclude()

        def f(*a, timeout=10, **k):
            try:
                # Can't serialize ports
                a = [i.name if isinstance(i, PortInfo) else i for i in a]
                x = self.rpc.call(attr, args=a, kwargs=k, block=0.001, timeout=timeout)
                if isinstance(x, dict):
                    x = PortInfo(**x)

                return x
            except TimeoutError:
                if timeout > 8:
                    print(traceback.format_exc())
                    self.worker.terminate()
                    self.worker.kill()
                    close_fds(self.worker)
                    workers.do(self.worker.wait)
                raise
            except Exception:
                print(traceback.format_exc())
                raise

        return f

    def get_all_connections(self, *a, **k):
        check_exclude()

        a = [i.name if isinstance(i, PortInfo) else i for i in a]
        x = self.rpc.call("get_all_connections", args=a, kwargs=k, block=0.001)
        x = [PortInfo(**i) for i in x]
        return x

    def get_ports(self, *a, **k):
        check_exclude()

        if self.ended or not self.worker.poll() is None:
            raise RuntimeError("This process is already dead")
        try:
            a = [i.name if isinstance(i, PortInfo) else i for i in a]
            x = self.rpc.call("get_ports", args=a, kwargs=k, block=0.001)
            x = [PortInfo(**i) for i in x]
            return x

        except TimeoutError:
            print(traceback.format_exc())
            self.worker.terminate()
            self.worker.kill()
            close_fds(self.worker)
            workers.do(self.worker.wait)
            raise

    def __del__(self):
        if self.worker:
            self.worker.terminate()
            self.worker.kill()
            close_fds(self.worker)
            workers.do(self.worker.wait)

    def on_port_registered(self, name, is_input, shortname, is_audio, registered):
        def f():
            try:
                global realConnections
                "Same function for register and unregister"
                # if not port:
                #     return

                p = PortInfo(name, is_input, shortname, is_audio)

                if registered:
                    # log.debug("JACK port registered: "+port.name)
                    with portsListLock:
                        portsList[name] = p
                    messagebus.post_message("/system/jack/newport", p)
                else:
                    torm = []
                    with portsListLock:
                        for i in _realConnections:
                            if i[0] == name or i[1] == name:
                                torm.append(i)
                        for i in torm:
                            del _realConnections[i]

                        try:
                            del portsList[name]
                        except Exception:
                            pass
                        realConnections = _realConnections.copy()

                    messagebus.post_message("/system/jack/delport", p)
            except Exception:
                print(traceback.format_exc())

        jackEventHandlingQueue.append(f)
        workers.do(handle_jack_event)

    def on_port_connected(self, a_is_output, a_name, b_name, connected):
        # Whem things are manually dis_connected we don't
        # Want to always reconnect every time
        if self.ended:
            return

        def f():
            global realConnections

            if connected:
                with portsListLock:
                    if a_is_output:
                        _realConnections[a_name, b_name] = True
                    else:
                        _realConnections[b_name, a_name] = True

                    realConnections = _realConnections.copy()

            if not connected:
                i = (a_name, b_name)
                with portsListLock:
                    if (a_name, b_name) in _realConnections:
                        try:
                            del _realConnections[a_name, b_name]
                        except KeyError:
                            pass

                    if (b_name, a_name) in _realConnections:
                        try:
                            del _realConnections[b_name, a_name]
                        except KeyError:
                            pass

                    realConnections = _realConnections.copy()

                # Try to stop whatever airwire or set therof
                # from remaking the connection
                if i in activeConnections:
                    try:
                        # Deactivate first, that must keep it from using the api
                        # From within the callback
                        activeConnections[i].active = False
                        del allConnections[i]
                        del activeConnections[i]
                    except Exception:
                        pass

                # def f():
                #     if not connected:
                #         log.debug("JACK port "+ a.name+" dis_connected from "+b.name)
                #     else:
                #         log.debug("JACK port "+ a.name+" connected to "+b.name)

                # workers.do(f)

        jackEventHandlingQueue.append(f)
        workers.do(handle_jack_event)

    def close(self):
        if self.ended:
            return

        self.ended = True
        if not self.worker.poll() is None:
            return
        try:
            x = self.rpc.call("close")
            self.rpc.stopFlag = True
        except Exception:
            self.rpc.stopFlag = True
            self.worker.terminate()
            self.worker.kill()
            close_fds(self.worker)
            workers.do(self.worker.wait)
            raise

    def __init__(self, *a, **k):
        # -*- coding: utf-8 -*-
        # If del can't find this it would to an infinite loop
        self.worker = None

        self.ended = False
        f = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "jack_client_subprocess.py"
        )
        env = {}
        env.update(os.environ)

        # Always use installed version.
        # TODO this will cause using the old one
        # if the new one isn't there, but is needed
        # for nixos et al compatibility
        if which("kaithem._jackmanager_server") and False:
            self.worker = Popen(
                ["kaithem._jackmanager_server"],
                stdout=PIPE,
                stdin=PIPE,
                stderr=STDOUT,
                env=env,
            )
        else:
            self.worker = Popen(
                ["python3", f], stdout=PIPE, stdin=PIPE, stderr=STDOUT, env=env
            )
        self.rpc = RPC(
            target=self, stdin=self.worker.stdout, stdout=self.worker.stdin, daemon=True
        )
        self.rpc.call("init")

    def print(self, s):
        print(s)


def handle_jack_event():
    with jackEventHandlingQueueLock:
        if jackEventHandlingQueue:
            f = jackEventHandlingQueue.pop(False)
            f()


dummy = False


log = logging.getLogger("system.jack")

_jackclient = None


lock = threading.RLock()


def on_jack_failure():
    pass


def on_jack_start():
    pass


# No other lock should ever be gotten under this, to ensure anti deadlock ordering.

# This also protects the list of connections.  There is a theoretical race condition currently,
# Siomeone else could disconnect a port right as we connect it, and we currently mark things connected by ourselves
# without waiting for jack to tell us, to avoid double connects if at all possible on the "Don't touch the scary server" principle.

# However, in basically all intended use cases there will never be any external things changing anything around, other than manual users who can quicky fix au
# misconnections.
portsListLock = threading.Lock()
portsList = {}


# Currently we only support using the default system card as the
# JACK backend. We prefer this because it's easy to get Pulse working right.
usingDefaultCard = True


def is_connected(f, t):
    if not isinstance(f, str):
        f = f.name
    if not isinstance(t, str):
        t = t.name

    if (t, f) in _realConnections:
        return True

    if (f, t) in _realConnections:
        return True


ensureConnectionsQueued = [0]


def _ensureConnections(*a, **k):
    "Auto restore connections in the connection list"

    # Someone else is already gonna run this
    # It is ok to have excess runs, but there must always be atleast 1 run after every change
    if ensureConnectionsQueued[0]:
        return
    ensureConnectionsQueued[0] = 1

    try:
        with lock:
            # Avoid race conditions, set flag BEFORE we check.
            # So we can't miss anything.  The other way would cause them to think we would check,
            # so they exit, but actually we already did.
            ensureConnectionsQueued[0] = 0
            x = list(allConnections.keys())
        for i in x:
            try:
                allConnections[i].reconnect()
            except KeyError:
                pass
            except Exception:
                print(traceback.format_exc())
    except Exception:
        ensureConnectionsQueued[0] = 0
        log.exception("Probably just a weakref that went away.")


def _checkNewAvailableConnection(*a, **k):
    "Auto restore connections in the connection list"
    try:
        with lock:
            x = list(allConnections.keys())
        for i in x:
            try:
                allConnections[i].reconnect()
            except Exception:
                print(traceback.format_exc())
    except Exception:
        log.exception("Probably just a weakref that went away.")


messagebus.subscribe("/system/jack/newport", _ensureConnections)

allConnections = weakref.WeakValueDictionary()

activeConnections = weakref.WeakValueDictionary()

# Things as they actually are
realConnections = {}

_realConnections = {}


def find_real():
    global realConnections, _realConnections
    with lock:
        p = _jackclient.get_ports(is_output=True)

        pl = {}
        for i in p:
            try:
                for j in _jackclient.get_all_connections(i):
                    pl[i.name, j.name] = True
            except Exception:
                log.exception("Err")
        with portsListLock:
            _realConnections = pl
            realConnections = _realConnections.copy()

            p = _jackclient.get_ports()
            # First time, get the initial list
            if not portsList:
                for i in p:
                    portsList[i.name] = i


errlog = []


latestAirWireForGivenPair = weakref.WeakValueDictionary()


class MonoAirwire:
    """Represents a connection that should always exist as long as there
    is a reference to this object. You can also enable and disable it with
    the connect() and disconnect() functions.

    They start out in the connected state
    """

    def __init__(self, orig, to):
        self.orig = orig
        self.to = to
        self.active = True

        if isinstance(orig, PortInfo):
            orig = orig.name
        if isinstance(to, PortInfo):
            to = to.name
        self.tupleid = (orig, to)
        latestAirWireForGivenPair[self.tupleid] = self

    def disconnect(self, force=True):
        global realConnections
        self.dis_connected = True
        try:
            del allConnections[self.orig, self.to]
        except Exception:
            pass

        if not force:
            # As garbage collection happens at uppredicatble times,
            # Don't disconnect if this airwire has been taken over by a new connection between the ports
            x = None
            try:
                x = latestAirWireForGivenPair[self.tupleid]
            except KeyError:
                pass

            if x and not x is self:
                return

        try:
            if lock.acquire(timeout=10):
                try:
                    if is_connected(self.orig, self.to):
                        disconnect(self.orig, self.to)
                        del activeConnections[self.orig, self.to]
                    try:
                        with portsListLock:
                            del _realConnections[self.orig, self.to]
                            realConnections = _realConnections.copy()
                    except KeyError:
                        pass
                    try:
                        with portsListLock:
                            del _realConnections[self.to, self.orig]
                            realConnections = _realConnections.copy()
                    except KeyError:
                        pass
                finally:
                    lock.release()
            else:
                raise RuntimeError("getting lock")

        except Exception:
            pass

    def __del__(self):
        # TODO: Is there any possible deadlock risk at all here?
        if self.active:
            self.disconnect(False)

    def connect(self):
        allConnections[self.orig, self.to] = self
        activeConnections[self.orig, self.to] = self

        self.connected = True
        self.reconnect()

    def reconnect(self):
        if (self.orig, self.to) in activeConnections:
            if self.orig and self.to:
                try:
                    if not is_connected(self.orig, self.to):
                        if lock.acquire(timeout=10):
                            try:
                                connect(self.orig, self.to)
                                with portsListLock:
                                    _realConnections[self.orig, self.to] = True
                                    global realConnections
                                    realConnections = _realConnections.copy()
                            finally:
                                lock.release()
                        else:
                            raise RuntimeError("Could not get lock")
                except Exception:
                    print(traceback.format_exc())


class MultichannelAirwire(MonoAirwire):
    "Link all outputs of f to all inputs of t, in sorted order"

    def _getEndpoints(self):
        f = self.orig
        if not f:
            return None, None

        t = self.to
        if not t:
            return None, None
        return f, t

    def reconnect(self):
        """Connects the outputs of channel strip(Or other JACK thing)  f to the inputs of t, one to one, until
        you run out of ports.

        Note that channel strips only have the main inputs but can have sends,
        so we have to distinguish them in the regex.
        """
        if not self.active:
            return
        f, t = self._getEndpoints()
        if not f:
            return
        f = f.replace("*:", "")
        t = t.replace("*:", "")

        if portsListLock.acquire(timeout=10):
            try:
                outPorts = sorted(
                    [
                        portsList[i]
                        for i in portsList
                        if i.split(":")[0] == f
                        and portsList[i].is_audio
                        and portsList[i].is_output
                    ],
                    key=lambda x: x.name,
                )
                inPorts = sorted(
                    [
                        portsList[i]
                        for i in portsList
                        if i.split(":")[0] == t
                        and portsList[i].is_audio
                        and (not portsList[i].is_output)
                    ],
                    key=lambda x: x.name,
                )
            finally:
                portsListLock.release()
        else:
            raise RuntimeError("Getting lock")

        # outPorts = _jackclient.get_ports(f+":*",is_output=True,is_audio=True)
        # inPorts = _jackclient.get_ports(t+":*",is_input=True,is_audio=True)
        # Connect all the ports
        for i in zip(outPorts, inPorts):
            if not is_connected(i[0].name, i[1].name):
                if lock.acquire(timeout=10):
                    try:
                        connect(i[0], i[1])
                        with portsListLock:
                            _realConnections[i[0].name, i[1].name] = True
                            realConnections = _realConnections.copy()
                    finally:
                        lock.release()
                else:
                    raise RuntimeError("Getting lock")

    def disconnect(self, force=True):
        global _jackclient
        check_exclude()

        if hasattr(self, "noNeedToDisconnect"):
            return

        f, t = self._getEndpoints()
        if not f:
            return

        if not force:
            # As garbage collection happens at uppredicatble times,
            # Don't disconnect if this airwire has been taken over by a new connection between the ports
            x = None
            try:
                x = latestAirWireForGivenPair[self.tupleid]
            except KeyError:
                pass

            if x and not x is self:
                return

        if portsListLock.acquire(timeout=10):
            try:
                outPorts = sorted(
                    [
                        portsList[i]
                        for i in portsList
                        if i.split(":")[0] == f
                        and portsList[i].is_audio
                        and portsList[i].is_output
                    ],
                    key=lambda x: x.name,
                )
                inPorts = sorted(
                    [
                        portsList[i]
                        for i in portsList
                        if i.split(":")[0] == t
                        and portsList[i].is_audio
                        and (not portsList[i].is_output)
                    ],
                    key=lambda x: x.name,
                )
            finally:
                portsListLock.release()

        if lock.acquire(timeout=10):
            try:
                # Connect all the ports
                for i in zip(outPorts, inPorts):
                    if is_connected(i[0], i[1]):
                        disconnect(i[0], i[1])
                        try:
                            del activeConnections[i[0].name, i[1].name]
                        except KeyError:
                            pass
            finally:
                lock.release()
        else:
            raise RuntimeError("getting lock")

    def __del__(self):
        workers.do(self.disconnect)


class CombiningAirwire(MultichannelAirwire):
    def reconnect(self):
        """Connects the outputs of channel strip f to the port t. As in all outputs
        to one input. If the destination is a client, connect all channnels of src to all of dest.
        """
        if not self.active:
            return
        f, t = self._getEndpoints()
        if not f:
            return
        if lock.acquire(timeout=10):
            try:
                if t.endswith("*"):
                    t = t[:-1]

                if f.endswith("*"):
                    f = f[:-1]

                if t.endswith(":"):
                    t = t[:-1]

                if f.endswith(":"):
                    f = f[:-1]

                outPorts = []
                inPorts = []
                with portsListLock:
                    for i in portsList:
                        if i.startswith(f + ":") or i == f:
                            if portsList[i].is_output and portsList[i].is_audio:
                                outPorts.append(i)
                        if i.split(":")[0] == t or i == t:
                            if portsList[i].is_input and portsList[i].is_audio:
                                inPorts.append(i)

                # Connect all the ports
                for i in outPorts:
                    for j in inPorts:
                        if not is_connected(i, j):
                            connect(i, j)

            finally:
                lock.release()

    def disconnect(self, force=False):
        f, t = self._getEndpoints()
        if not f:
            return

        if not force:
            # As garbage collection happens at uppredicatble times,
            # Don't disconnect if this airwire has been taken over by a new connection between the ports
            x = None
            try:
                x = latestAirWireForGivenPair[self.tupleid]
            except KeyError:
                pass

            if x and not x is self:
                return

        if lock.acquire(timeout=10):
            try:
                if t.endswith("*"):
                    t = t[:-1]

                if f.endswith("*"):
                    f = f[:-1]

                outPorts = []
                inPorts = []
                with portsListLock:
                    for i in portsList:
                        if i.startswith(f + ":") or i == f:
                            if portsList[i].is_output and portsList[i].is_audio:
                                outPorts.append(i)
                        if i.split(":")[0] == t or i == t:
                            if portsList[i].is_input and portsList[i].is_audio:
                                inPorts.append(i)

                if not inPorts:
                    return
                # Disconnect all the ports
                for i in outPorts:
                    for j in inPorts:
                        if is_connected(i, j):
                            try:
                                disconnect(i, j)
                            except Exception:
                                print(traceback.format_exc())
                            try:
                                del activeConnections[i, j]
                            except KeyError:
                                pass
            finally:
                lock.release()


def Airwire(f, t, force_combining=False):
    # Can't connect to nothing, for now lets use a hack and make these nonsense
    # names so emoty strings don't connect to stuff
    if not f or not t:
        f = "jdgdsjfgkldsf"
        t = "dsfjgjdsfjgkl"
    if force_combining:
        return CombiningAirwire(f, t)
    elif f == None or t == None:
        return MonoAirwire(None, None)
    elif ":" in f:
        if not ":" in t:
            return CombiningAirwire(f, t)
        return MonoAirwire(f, t)
    else:
        return MultichannelAirwire(f, t)


############################################################################
# This section manages the actual sound IO and creates jack ports
# This code runs once when the event loads. It also runs when you save the event during the test compile
# and may run multiple times when kaithem boots due to dependancy resolution
__doc__ = ""


def work():
    global _reconnecterThreadObjectStopper

    # Wait 10s before actually doing anything to avoid nuisiance chattering errors.
    # This thread mostly only fixes crashed stuff.
    for i in range(100):
        if not _reconnecterThreadObjectStopper[0]:
            return
        time.sleep(0.1)

    failcounter = 0
    while _reconnecterThreadObjectStopper[0]:
        try:
            # The _checkJack stuf won't block, because we already have the lock
            if lock.acquire(timeout=2):
                failcounter = 0
                try:
                    _checkJackClient()
                finally:
                    lock.release()
                # _ensureConnections()
            else:
                if _reconnecterThreadObjectStopper[0]:
                    raise RuntimeError("Could not get lock,retrying in 5s")

                else:
                    # Already stopping anyway, ignore
                    pass
            time.sleep(5)
        except Exception:
            time.sleep(30)
            logging.exception("Error in jack manager")


_reconnecterThreadObject = None
_reconnecterThreadObjectStopper = [0]


def start_managing(p=None, n=None):
    "Start mananaging JACK in whatever way was configured."

    global _jackclient
    global _reconnecterThreadObject

    with lock:
        try:
            _jackclient = JackClientProxy()
        except Exception:
            log.exception("Error creating JACK client, retry later")

        try:
            find_real()
        except Exception:
            log.exception("Error getting initial jack graph info")

        # Stop the old thread if needed
        _reconnecterThreadObjectStopper[0] = 0
        try:
            if _reconnecterThreadObject:
                _reconnecterThreadObject.join()
        except Exception:
            pass

        _reconnecterThreadObjectStopper[0] = 1
        _reconnecterThreadObject = threading.Thread(target=work)
        _reconnecterThreadObject.name = "JackReconnector"
        _reconnecterThreadObject.daemon = True
        _reconnecterThreadObject.start()


def stop_managing():
    with lock:
        try:
            if _jackclient:
                _jackclient.exit()
        except Exception:
            pass
        # Stop the old thread if needed
        _reconnecterThreadObjectStopper[0] = 0
        try:
            if _reconnecterThreadObject:
                _reconnecterThreadObject.join()
        except Exception:
            pass
        _reconnecterThreadObject = None


postedCheck = True

firstConnect = False


def _checkJackClient(err=True):
    global _jackclient, realConnections, postedCheck, firstConnect
    import jack

    if lock.acquire(timeout=10):
        try:
            t = _jackclient.get_ports()

            if not t:
                if firstConnect:
                    raise RuntimeError(
                        "JACK Server not started or client not connected, will try connect "
                    )
                firstConnect = True

            if not postedCheck:
                postedCheck = True

            return True
        except Exception:
            postedCheck = False

            if firstConnect:
                firstConnect = True
            else:
                print(traceback.format_exc())

            try:
                _jackclient.close()
                _jackclient = None
            except Exception:
                pass

            with portsListLock:
                portsList.clear()
                _realConnections = {}

            try:
                _jackclient = JackClientProxy()
            except Exception:
                if err:
                    log.exception("Error creating JACK client")
                return

            _jackclient.get_ports()
            get_ports()
            time.sleep(0.5)
            find_real()
            return True
        finally:
            lock.release()

    if not _jackclient:
        return False


def get_portsListCache():
    "We really should not need to have this refreser, it is only there in case of erro, hence the 1 hour."
    global portsList, portsCacheTime
    if time.monotonic() - portsCacheTime < 3600:
        return portsList
    portsCacheTime = time.monotonic()
    get_ports()
    return portsList


portsCacheTime = 0


lastCheckedClientFromget_ports = 0


def get_ports(*a, max_wait=10, **k):
    global portsList, _jackclient, lastCheckedClientFromget_ports

    if lock.acquire(timeout=max_wait):
        try:
            if not _jackclient:
                # MOstly here so we can use this standalone from a unit test
                if lastCheckedClientFromget_ports < time.monotonic() - 120:
                    lastCheckedClientFromget_ports = time.monotonic()
                    workers.do(_checkJackClient)
                return []
            ports = []
            x = _jackclient.get_ports(*a, **k)

            with portsListLock:
                # No filters means this must be the full list
                if not a and not k:
                    portsList.clear()
                for port in x:
                    portsList[port.name] = port

            return x
        finally:
            lock.release()
    return []


def get_port_names_with_aliases(*a, **k):
    if lock.acquire(timeout=10):
        try:
            if not _jackclient:
                return []
            ports = []
            x = _jackclient.get_ports(*a, **k)
            for i in x:
                for j in i.aliases:
                    if not j in ports:
                        ports.append(j)
                if not i.name in ports:
                    ports.append(i.name)
            return ports
        finally:
            lock.release()
    else:
        pass


def get_connections(name, *a, **k):
    if lock.acquire(timeout=10):
        try:
            if not _jackclient:
                return []
            try:
                return _jackclient.get_all_connections(name)
            except Exception:
                log.exception("Error getting connections")
                return []
        finally:
            lock.release()
    else:
        pass


exclude_until = [0]


def check_exclude():
    if time.monotonic() < exclude_until[0]:
        raise RuntimeError("That is not allowed, trying to auto-fix")


def disconnect(f, t):
    global realConnections
    if lock.acquire(timeout=30):
        try:
            if not is_connected(f, t):
                return

            try:
                if isinstance(f, PortInfo):
                    f = f.name
                if isinstance(t, PortInfo):
                    t = t.name

                # Horrid hack to keep dummy connections around to not make gst stop
                if "SILENCE" in f:
                    return

                # This feels race conditionful but i think it is important so that we don't try to double-disconnect.
                # Be defensive with jack, the whole thing seems britttle
                # Let other side handle figuring out which is which
                for i in range(24):
                    # For unknown reasons it is possible to completely clog up the jack client.
                    # We must make a new one and retry should this ever happen
                    try:
                        _jackclient.disconnect(f, t, timeout=5)
                        # subprocess.check_call(['pw-jack', 'jack_disconnect', f, t])
                        break
                    except TimeoutError:
                        if (i % 6) == 5:
                            time.sleep(5)
                        _jackclient.worker.kill()
                        close_fds(_jackclient.worker)
                        _checkJackClient()

                with portsListLock:
                    try:
                        del _realConnections[f, t]
                        realConnections = _realConnections.copy()
                    except KeyError:
                        pass

                    try:
                        del _realConnections[t, f]
                        realConnections = _realConnections.copy()
                    except KeyError:
                        pass

            except Exception:
                print(traceback.format_exc())
        finally:
            lock.release()
    else:
        pass


def disconnect_all_from(p: str):
    "Disconnect everything to do wth given port"
    find_real()
    for i in list(realConnections.keys()):
        if i[0] == p or i[0].startswith(p + ":"):
            disconnect(*i)
        elif i[1] == p or i[1].startswith(p + ":"):
            disconnect(*i)


# This is an easy place to get a bazillion sounds queued up all waiting on the lock. This stops that.
awaiting = [0]
awaitingLock = threading.Lock()


def connect(f, t, ts=None):
    ts = ts or time.monotonic()

    global realConnections, _jackclient
    check_exclude()
    with awaitingLock:
        if awaiting[0] > 8:
            time.sleep(1)

        if awaiting[0] > 12:
            raise RuntimeError("Too many threads are waiting to make JACK connections")

        awaiting[0] += 1

    try:
        if lock.acquire(timeout=10):
            try:
                if is_connected(f, t):
                    return

                try:
                    if isinstance(f, PortInfo):
                        f = f.name
                    if isinstance(t, PortInfo):
                        t = t.name
                except Exception:
                    return

                try:
                    # Let other side handle figuring out which is which
                    for i in range(3):
                        # For unknown reasons it is possible to completely clog up the jack client.
                        # We must make a new one and retry should this ever happen
                        try:
                            _jackclient.connect(t, f, timeout=10)
                            break
                        except TimeoutError:
                            _jackclient.worker.kill()
                            close_fds(_jackclient.worker)
                            _checkJackClient()
                    with portsListLock:
                        try:
                            _realConnections[f, t] = True
                            realConnections = _realConnections.copy()
                        except KeyError:
                            pass
                except Exception:
                    print(traceback.format_exc())
            finally:
                lock.release()
        else:
            pass
    finally:
        with awaitingLock:
            awaiting[0] -= 1



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

from __future__ import annotations
import time
import sys
import functools
import base64
import os
import threading
import weakref
from typing import Optional
from subprocess import PIPE, STDOUT
from subprocess import Popen
from scullery import workers
from .jsonrpyc import RPC


def close_fds(p: Popen):
    try:
        p.stdin.close()
    except Exception:
        pass
    try:
        p.stdout.close()
    except Exception:
        pass
    try:
        p.stderr.close()
    except Exception:
        pass


@functools.cache
def which(program):
    "Check if a program is installed like you would do with UNIX's which command."

    # Because in windows, the actual executable name has .exe while the command name does not.
    if sys.platform == "win32" and not program.endswith(".exe"):
        program += ".exe"

    # Find out if path represents a file that the current user can execute.
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    # If the input was a direct path to an executable, return it
    if fpath:
        if is_exe(program):
            return program

    # Else search the path for the file.
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    # If we got this far in execution, we assume the file is not there and return None
    return None


# Can't pass GST elements, have to pass IDs
class ElementProxy:
    "Local proxy to element in the server process"

    def __init__(self, parent: GstreamerPipeline, obj_id) -> None:
        # This was making a bad GC loop issue.
        self.parent = weakref.ref(parent)
        self.id = obj_id

    def set_property(self, p, v, max_wait=10):
        x = self.parent()
        assert x
        x.set_property(self.id, p, v, max_wait=max_wait)

    def pull_buffer(self, timeout=0.1):
        x = self.parent()
        assert x
        return base64.b64decode(x.pull_buffer(self.id, timeout))

    def pull_to_file(self, f):
        x = self.parent()
        assert x
        return x.pull_to_file(self.id, f)


pipes = weakref.WeakValueDictionary()


class GStreamerPipeline:
    def __init__(self, *a, **k):
        # -*- coding: utf-8 -*-

        # If del can't find this it would to an infinite loop
        self.worker: Optional[Popen] = None

        pipes[id(self)] = self
        self.ended = False
        f = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "iceflow_server.py"
        )
        # Unusued, the lock is for compatibility wiith the old not-rpc based iceflow
        self.lock = threading.RLock()
        env = {}
        env.update(os.environ)
        env["GST_DEBUG"] = "*:1"

        self.rpc = None

        if which("kaithem._iceflow_server") and False:
            self.worker = Popen(
                ["kaithem._iceflow_server"],
                stdout=PIPE,
                stdin=PIPE,
                stderr=STDOUT,
                env=env,
            )
        else:
            # TODO nobody seems to know why this sometimes OSErrors
            for i in range(5):
                try:
                    self.worker = Popen(
                        ["python3", f], stdout=PIPE, stdin=PIPE, stderr=STDOUT, env=env
                    )
                    self.worker.stdin.flush()
                    break
                except OSError:
                    if i == 4:
                        raise
                    else:
                        time.sleep(i / 10)

        self.rpc = RPC(
            target=weakref.proxy(self),
            stdin=self.worker.stdout,
            stdout=self.worker.stdin,
            daemon=True,
        )
        # We have no way of knowing when it's actually ready and listening for commands if gstreamer
        # needs to load
        time.sleep(1)

    def rpc_call(self, *a, **k):
        if self.rpc:
            return self.rpc.call(*a, **k)

        raise RuntimeError("No RPC object")

    def __getattr__(self, attr):
        if self.ended or self.worker.poll() is not None:
            raise RuntimeError("This process is already dead")

        def f(*a, **k):
            try:
                return self.rpc_call(attr, args=a, kwargs=k, block=0.001, timeout=15)
            except Exception:
                if self.worker:
                    self.cleanup_popen()
                    workers.do(self.worker.wait)
                raise

        return f

    def pull_to_file(self, *a, **k):
        if self.ended or self.worker.poll() is not None:
            raise RuntimeError("This process is already dead")

        try:
            return self.rpc_call(
                "pull_to_file", args=a, kwargs=k, block=0.001, timeout=0.5
            )
        except Exception:
            if self.worker:
                self.cleanup_popen()
                workers.do(self.worker.wait)
            raise

    def cleanup_popen(self):
        self.worker.terminate()
        self.worker.kill()
        close_fds(self.worker)
        try:
            self.rpc.stdin.close()
        except:
            pass

    def __del__(self):
        self.cleanup_popen()
        workers.do(self.worker.wait)

    def add_element(self, element_name: str, *a, **k):
        "Returns an element proxy object"
        # This has to do with setup and I suppose we probably shouldn't just let the error pass by.
        if self.ended or self.worker.poll() is not None:
            raise RuntimeError("This process is already dead")

        # convert element proxies to their ids for transmission
        for key, item in k.items():
            if isinstance(item, ElementProxy):
                k[key] = item.id

        if "connectToOutput" in k:
            k["connect_to_output"] = k.pop("connectToOutput")

        if "connect_to_output" in k and isinstance(
            k["connect_to_output"], (list, tuple)
        ):
            k["connect_to_output"] = [
                (i.id if isinstance(i, ElementProxy) else i)
                for i in k["connect_to_output"]
            ]
        return ElementProxy(
            self,
            self.rpc_call(
                "add_elementRemote",
                args=(element_name, *a),
                kwargs=k,
                block=0.0001,
                timeout=5,
            ),
        )

    def set_property(self, *a, max_wait=10, **k):
        # Probably Just Not Important enough to raise an error for this.
        if self.ended or self.worker.poll() is not None:
            print("Prop set in dead process")
            self.ended = True
            return

        # convert element proxies to their ids for transmission
        for key, item in k.items():
            if isinstance(item, ElementProxy):
                k[key] = item.id

        a = [i.id if isinstance(i, ElementProxy) else i for i in a]
        return ElementProxy(
            self,
            self.rpc_call(
                "set_property", args=a, kwargs=k, block=0.0001, timeout=max_wait
            ),
        )

    def add_pil_capture(self, *a, **k):
        # Probably Just Not Important enough to raise an error for this.
        if self.ended or self.worker.poll() is not None:
            print("Prop set in dead process")
            self.ended = True
            return
        return ElementProxy(
            self,
            self.rpc_call(
                "addRemotePILCapture", args=a, kwargs=k, block=0.0001, timeout=10
            ),
        )

    def on_appsink_data(self, element_name, data, *a, **k):
        return

    def _on_appsink_data(self, element_name, data):
        self.on_appsink_data(element_name, base64.b64decode(data))

    def on_motion_begin(self, *a, **k):
        print("Motion start")

    def on_motion_end(self, *a, **k):
        print("Motion end")

    def on_multi_file_sink_file(self, fn, *a, **k):
        print("MultiFileSink", fn)

    def on_barcode(self, codetype, data):
        print("Barcode: ", codetype, data)

    def on_level_message(self, src, rms, level):
        pass

    def stop(self):
        if self.ended:
            return

        self.ended = True
        if self.worker.poll() is not None:
            self.ended = True
            return
        try:
            self.rpc_call("stop", block=0.01, timeout=10)
            self.worker.terminate()
            time.sleep(0.5)
            self.worker.kill()
            close_fds(self.worker)
            try:
                self.rpc.stdin.close()
            except:
                pass

        except Exception:
            self.worker.terminate()
            time.sleep(0.5)
            self.worker.kill()
            close_fds(self.worker)
            try:
                self.rpc.stdin.close()
            except:
                pass
            workers.do(self.worker.wait)

        self.rpc = None

    def add_jack_mixer_send_elements(self, *a, **k):
        a, b = self.rpc_call(
            "add_jack_mixer_send_elements", args=a, kwargs=k, block=0.0001, timeout=10
        )
        return (ElementProxy(self, a), ElementProxy(self, b))

    def print(self, s):
        print(s)

    def on_presence_value(self, v):
        print(v)


GstreamerPipeline = GStreamerPipeline



import gc
import random
import unittest
import time
import os
import icemedia.iceflow
import weakref

testmedia = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "Brothers Unite.opus"
)


class Player(icemedia.iceflow.GstreamerPipeline):
    def __init__(self, file):
        icemedia.iceflow.GstreamerPipeline.__init__(self, realtime=False)

        self.src = self.add_element("filesrc", location=file)

        # This bin autodetects and decodes basically any type of media
        # It is special cased, anything onnected to it is actually connected on-demand as needed
        decodebin = self.add_element("decodebin")

        self.add_element("audioconvert", connectToOutput=decodebin)
        self.add_element("audioresample")

        self.fader = self.add_element("volume", volume=1)
        self.sink = self.add_element("autoaudiosink")


class TestAudio(unittest.TestCase):
    def test_z_no_segfaults(self):
        # Test for segfault-ery
        for i in range(100):
            p = Player(testmedia)
            p.start()
            time.sleep(3 * random.random())
            p.seek(0.3)
            time.sleep(0.01 * random.random())
            p.set_property(p.fader, "volume", 0.1)
            p.stop()
            # Ensure nothing bad happens setting the volume after stopping
            p.set_property(p.fader, "volume", 1)
            del p
            gc.collect()
        for i in range(150):
            time.sleep(0.1)
            if len(icemedia.iceflow.pipes) == 0:
                break
        gc.collect()

        self.assertEqual(len(icemedia.iceflow.pipes), 0)

    def test_play(self):
        # Test for segfault-ery
        for i in range(2):
            p = Player(testmedia)
            p.start()
            time.sleep(2 * random.random())
            p.seek(0.3)
            time.sleep(0.01 * random.random())
            p.set_property(p.fader, "volume", 0.1)
            time.sleep(3)
            p.stop()
            # Ensure nothing bad happens setting the volume after stopping
            p.set_property(p.fader, "volume", 1)
            p2 = weakref.ref(p)
            del p
            time.sleep(3)
            assert not p2()
            gc.collect()

        for i in range(150):
            time.sleep(0.1)
            if len(icemedia.iceflow.pipes) == 0:
                break
        gc.collect()

        self.assertEqual(len(icemedia.iceflow.pipes), 0)

    def test_seekpastend(self):
        p = Player(testmedia)
        p.start()
        time.sleep(0.01 * random.random())
        p.seek(99999)
        time.sleep(0.01 * random.random())
        p.set_property(p.fader, "volume", 1)
        p.stop()
        # Ensure nothing bad happens setting the volume after stopping
        p.set_property(p.fader, "volume", 1)
        del p
        gc.collect()
        for i in range(150):
            time.sleep(0.1)
            if len(icemedia.iceflow.pipes) == 0:
                break
        gc.collect()

        self.assertEqual(len(icemedia.iceflow.pipes), 0)



# SPDX-FileCopyrightText: Copyright 2013 Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

import subprocess

import os
import time
import threading
import collections
import logging
import traceback
from typing import List, Any, Optional, Callable

from scullery import workers

from scullery import util

from .python_mpv_jsonipc import MPV

log = logging.getLogger("system.sound")


def tryCloseFds(p: Optional[subprocess.Popen[Any]]):
    if p:
        try:
            if p.stdout:
                p.stdout.close()
        except Exception:
            pass
        try:
            if p.stderr:
                p.stderr.close()
        except Exception:
            pass
        try:
            if p.stdin:
                p.stdin.close()
        except Exception:
            pass


media_paths = [""]

media_resolvers: dict[str, Callable[[str], str | None]] = {}


def resolve_sound(fn: str, extrapaths: List[str] = []) -> str:
    "Get the full path of a sound file by searching"
    filename = util.search_paths(fn, extrapaths)
    if not filename:
        filename = util.search_paths(fn, media_paths)

    # Search all module media folders
    if not filename:
        for k, i in media_resolvers.items():
            filename = i(fn)
            if filename:
                break

    # Raise an error if the file doesn't exist
    if not filename or not os.path.isfile(filename):
        raise ValueError("Specified audio file '" + fn + "' was not found")
    assert isinstance(filename, str)
    return filename


# This class provides some infrastructure to play sounds but if you use it directly it is a dummy.
class SoundWrapper(object):
    backendname = "Dummy Sound Driver(No real sound player found)"

    def __init__(self):
        # Prefetch cache for preloadng sound effects
        self.cache = collections.OrderedDict()
        self.runningSounds = {}

    def readySound(self, *args, **kwargs):
        pass

    @staticmethod
    def testAvailable():
        # Default to command based test
        return False

    # little known fact: Kaithem is actually a large collection of
    # mini garbage collectors and bookkeeping code...
    def delete_stopped_sounds(self):
        x = list(self.runningSounds.keys())
        for i in x:
            try:
                if not self.runningSounds[i].is_playing():
                    self.runningSounds.pop(i)
            except KeyError:
                pass

    def stop_all_sounds(self):
        x = list(self.runningSounds.keys())
        for i in x:
            try:
                self.runningSounds.pop(i)
            except KeyError:
                pass

    def get_position(self, channel: str = "PRIMARY"):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].position()
        except KeyError:
            return False

    def set_volume(self, vol: float, channel: str = "PRIMARY"):
        pass

    def set_speed(self, speed: float, channel: str = "PRIMARY", *a, **kw):
        pass

    def play_sound(
        self,
        filename: str,
        handle: str = "PRIMARY",
        extraPaths: List[str] = [],
        volume: float = 1,
        finalGain: Optional[float] = None,
        output: Optional[str] = "",
        loop: float = 1,
        start: float = 0,
        speed: float = 1,
    ):
        pass

    def stop_sound(self, handle: str = "PRIMARY"):
        pass

    def is_playing(self, handle: str = "blah", refresh: bool = False):
        return False

    def pause(self, handle: str = "PRIMARY"):
        pass

    def resume(self, handle: str = "PRIMARY"):
        pass

    def fade_to(
        self,
        file: str | None,
        length: float = 1.0,
        block: bool = False,
        handle: str = "PRIMARY",
        output: Optional[str] = "",
        volume: float = 1,
        windup: float = 0,
        winddown: float = 0,
        loop: int = 1,
        start: float = 0,
        speed: float = 1,
    ):
        if file:
            self.play_sound(file, handle)
        else:
            self.stop_sound(handle)

    def preload(self, filename: str):
        pass

    def seek(self, position: float, channel: str = "PRIMARY"):
        try:
            return self.runningSounds[channel].seek(position)
        except KeyError:
            pass


test_sound_logs = []
play_logs = []


class TestSoundWrapper(SoundWrapper):
    def play_sound(
        self,
        filename: str,
        handle: str = "PRIMARY",
        extraPaths: List[str] = [],
        volume: float = 1,
        finalGain: float | None = None,
        output: str | None = "",
        loop: float = 1,
        start: float = 0,
        speed: float = 1,
    ):
        test_sound_logs.append(["play", handle, filename])
        play_logs.append(["play", handle, filename])

    def fade_to(
        self,
        file: str | None,
        length: float = 1,
        block: bool = False,
        handle: str = "PRIMARY",
        output: str | None = "",
        volume: float = 1,
        windup: float = 0,
        winddown: float = 0,
        loop: int = 1,
        start: float = 0,
        speed: float = 1,
    ):
        test_sound_logs.append(["fade_to", handle, file])
        play_logs.append(["fade_to", handle, file])

    def stop_sound(self, handle: str = "PRIMARY"):
        test_sound_logs.append(["stop", handle])


objectPoolLock = threading.RLock()

objectPool = []


class PlayerHolder(object):
    def __init__(self, p: MPV) -> None:
        self.player = p
        self.usesCounter = 0
        self.conf = [0]
        self.isConfigured = False
        self.lastvol = -99089798
        self.conf_speed = 1
        self.loop_conf = -1
        self.alreadyMadeReplacement = False
        self.lastjack = "hgfdxcghjkufdszcxghjkuyfgdx"


class MPVBackend(SoundWrapper):
    @staticmethod
    def testAvailable():
        if not util.which("mpv"):
            return False
        try:
            return True
        except Exception:
            pass

    backendname = "MPV"

    # What this does is it keeps a reference to the sound player process and
    # If the object is destroyed it destroys the process stopping the sound
    # It also abstracts checking if its playing or not.
    class MPVSoundContainer(object):
        def __init__(
            self,
            filename,
            vol,
            finalGain,
            output,
            loop,
            start=0.0,
            speed=1.0,
            just_preload=False,
        ):
            self.lock = threading.RLock()
            self.stopped = False
            self.is_playingCache = None

            self.player: Optional[PlayerHolder] = None

            if output == "__disable__":
                return

            self.alreadySetCorrection = False

            # I think this leaks memory when created and destroyed repeatedly
            with objectPoolLock:
                if len(objectPool):
                    self.player = objectPool.pop()
                else:
                    self.player = PlayerHolder(MPV())

            # Avoid somewhat slow RPC calls if we can
            if not self.player.isConfigured:
                cname = "kplayer" + str(time.monotonic()) + "_out"
                self.player.player.vid = "no"
                self.player.player.keep_open = "yes"
                self.player.player.ao = "jack,pulse,alsa"
                self.player.player.jack_name = cname
                self.player.player.gapless_audio = "weak"
                self.player.player.isConfigured = True

            if speed != self.player.conf_speed:
                self.player.player.audio_pitch_correction = False
                self.player.player.speed = speed

            if not loop == self.player.loop_conf:
                # For legavy reasons some stuff used tens of millions instead of actual infinite loop.
                # But it seems mpv ignores everything past a certain number. So we replace effectively forever with
                # actually forever to get the same effect, assuming that was user intent.
                if not (loop == -1 or loop > 900000000):
                    self.player.player.loop_playlist = int(max(loop, 1))
                else:
                    self.player.player.loop_playlist = "inf"

            # Due to memory leaks, these have a limited lifespan
            self.player.usesCounter += 1

            if (not hasattr(self.player, "lastvol")) or not self.player.lastvol == vol:
                self.player.lastvol = vol
                self.player.player.volume = vol * 100

            self.volume = vol
            self.finalGain = finalGain if finalGain is not None else vol

            jp = "system:*"
            if output:
                if ":" not in output:
                    jp = output + ":*"
                else:
                    jp = output

            if not self.player.lastjack == jp:
                self.player.player.jack_port = jp
                self.player.player.lastjack = jp

            self.started = time.time()

            if filename:
                if self.player:
                    self.is_playingCache = None
                    self.player.player.loadfile(filename)

                    self.player.player.pause = False
                    if start:
                        for i in range(50):
                            try:
                                time.sleep(0.01)
                                self.player.player.seek(str(start), "absolute")
                                break
                            except Exception:
                                pass

                else:
                    raise RuntimeError("Player object is gone")

        def __del__(self):
            self.stop()

        def stop(self):
            if self.stopped:
                return
            self.stopped = True
            bad = True
            if self.player:
                # Only do the maybe recycle logic when stopping a still good SFX

                try:
                    with self.lock:
                        self.player.player.stop()
                    bad = False
                except Exception:
                    # Sometimes two threads try to stop this at the same time and we get a race condition
                    # I really hate this but stopping a crashed sound can't be allowed to take down anything else.
                    pass

                # When the player only has a few uses left, if we don't have many spare objects in
                # the pool, we are going to make the replacement ahead of time in a background thread.
                # But try tpo only make one replacement per object, we don't actually want to go up to the max
                # in the pool because they can use CPU in the background
                if bad or self.player.usesCounter > 8:
                    if not self.player.alreadyMadeReplacement:
                        if (len(objectPool) < 3) or self.player.usesCounter > 10:
                            self.player.alreadyMadeReplacement = True

                            def f():
                                # Can't make it under lock that is slow
                                from .python_mpv_jsonipc import MPV

                                o = PlayerHolder(MPV())
                                with objectPoolLock:
                                    if len(objectPool) < 4:
                                        objectPool.append(o)
                                        return
                                o.player.stop()

                            workers.do(f)

                if bad or self.player.usesCounter > 10:
                    p = self.player

                    def f():
                        if p:
                            p.player.stop()

                    workers.do(f)

                else:
                    with objectPoolLock:
                        p = self.player
                        if p:
                            if len(objectPool) < 4:
                                objectPool.append(p)
                            else:
                                self.player.player.stop()
                self.player = None

        def is_playing(self, refresh=False):
            with self.lock:
                if not self.player:
                    return False
                try:
                    if not refresh:
                        if self.is_playingCache is not None:
                            return self.is_playingCache
                    c = self.player.player.eof_reached == False

                    if c is False:
                        self.is_playingCache = c

                    return c
                except Exception:
                    logging.exception("Error getting playing status, assuming closed")
                    return False

        def position(self):
            return time.time() - self.started

        def wait(self):
            with self.lock:
                if self.player:
                    self.player.wait_for_playback()
                else:
                    raise RuntimeError("No player object")

        def seek(self, position):
            pass

        def setVol(self, volume, final=True):
            with self.lock:
                self.volume = volume
                if final:
                    self.finalGain = volume
                if self.player:
                    self.player.lastvol = volume
                    self.player.player.volume = volume * 100

        def set_speed(self, speed):
            with self.lock:
                if self.player:
                    if not self.alreadySetCorrection:
                        self.player.player.audio_pitch_correction = False
                        self.alreadySetCorrection = True
                    # Mark as needing to be updated
                    self.player.conf_speed = speed
                    self.player.player.speed = speed

        def getVol(self):
            with self.lock:
                if self.player:
                    return self.player.player.volume
                else:
                    return 0

        def pause(self):
            with self.lock:
                if self.player:
                    self.player.player.pause = True

        def resume(self):
            with self.lock:
                if self.player:
                    self.player.player.pause = False

    def play_sound(
        self,
        filename: str,
        handle: str = "PRIMARY",
        extraPaths: List[str] = [],
        volume: float = 1,
        finalGain: Optional[float] = None,
        output: Optional[str] = "",
        loop: float = 1,
        start: float = 0,
        speed: float = 1,
    ):
        # Those old sound handles won't garbage collect themselves
        self.delete_stopped_sounds()
        # Raise an error if the file doesn't exist
        fn = resolve_sound(filename, extraPaths)
        # Play the sound with a background process and keep a reference to it
        self.runningSounds[handle] = self.MPVSoundContainer(
            fn, volume, finalGain, output, loop, start=start, speed=speed
        )

    def stop_sound(self, handle="PRIMARY"):
        # Delete the sound player reference object and its destructor will stop the sound
        if handle in self.runningSounds:
            # Instead of using a lock lets just catch the error is someone else got there first.
            try:
                x = self.runningSounds[handle]
                try:
                    x.stop()
                except Exception:
                    logging.exception("Error stopping")
                del self.runningSounds[handle]
                x.nocallback = True
                del x
            except KeyError:
                pass

    def is_playing(self, channel="PRIMARY", refresh=False):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].is_playing(refresh)
        except KeyError:
            return False

    def wait(self, channel="PRIMARY"):
        "Block until any sound playing on a channel is finished"
        try:
            self.runningSounds[channel].wait()
        except KeyError:
            return False

    def set_volume(self, vol, channel="PRIMARY", final=True):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].setVol(vol, final=final)
        except KeyError:
            pass

    def set_speed(self, speed, channel="PRIMARY", *a, **kw):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].set_speed(speed)
        except KeyError:
            pass

    def seek(self, position, channel="PRIMARY"):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].seek(position)
        except KeyError:
            pass

    def pause(self, channel="PRIMARY"):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].pause()
        except KeyError:
            pass

    def resume(self, channel="PRIMARY"):
        "Return true if a sound is playing on channel"
        try:
            return self.runningSounds[channel].resume()
        except KeyError:
            pass

    def fade_to(
        self,
        file: str | None,
        length: float = 1.0,
        block: bool = False,
        handle: str = "PRIMARY",
        output: Optional[str] = "",
        volume: float = 1,
        windup: float = 0,
        winddown: float = 0,
        loop: int = 1,
        start: float = 0,
        speed: float = 1,
    ):
        x = self.runningSounds.pop(handle, None)

        if x and not (length or winddown):
            x.stop()

        # Allow fading to silence
        if file:
            sspeed = speed
            if windup:
                sspeed = 0.1

            self.play_sound(
                file,
                handle=handle,
                volume=0,
                output=output,
                finalGain=volume,
                loop=loop,
                start=start,
                speed=sspeed,
            )

        # if not x:
        #    return
        if not (length or winddown or windup):
            return

        def f():
            t = time.monotonic()
            try:
                v = x.volume
            except Exception:
                v = 0

            targetVol = 1
            while time.monotonic() - t < max(length, winddown, windup):
                if max(length, winddown):
                    foratio = max(
                        0, min(1, ((time.monotonic() - t) / max(length, winddown)))
                    )
                else:
                    foratio = 1

                if length:
                    firatio = max(0, min(1, ((time.monotonic() - t) / length)))
                else:
                    firatio = 1

                tr = time.monotonic()

                if x and x.player:
                    # Player might have gotten itself stopped by now
                    try:
                        x.setVol(max(0, v * (1 - foratio)))

                        if winddown:
                            wdratio = max(
                                0, min(1, ((time.monotonic() - t) / winddown))
                            )
                            x.set_speed(max(0.1, speed * (1 - wdratio)))
                    except AttributeError:
                        print(traceback.format_exc())

                if file and (handle in self.runningSounds):
                    targetVol = self.runningSounds[handle].finalGain
                    self.set_volume(min(1, targetVol * firatio), handle, final=False)

                    if windup:
                        wuratio = max(0, min(1, ((time.monotonic() - t) / windup)))
                        self.set_speed(max(0.1, min(speed, wuratio * speed, 8)), handle)

                # Don't overwhelm the backend with commands
                time.sleep(max(1 / 48.0, time.monotonic() - tr))

            try:
                targetVol = self.runningSounds[handle].finalGain
            except KeyError:
                targetVol = -1

            if not targetVol == -1:
                try:
                    self.set_volume(min(1, targetVol), handle)
                except Exception as e:
                    print(e)

            if x:
                x.stop()

        if block:
            f()
        else:
            workers.do(f)


backends = {"mpv": MPVBackend, "test": TestSoundWrapper}

backend = SoundWrapper()


def select_backend(name: str):
    global backend
    global play_sound, stop_sound, is_playing, resolve_sound, pause, resume
    global fade_to, position, setvol, readySound, preload

    if name not in backends:
        raise RuntimeError("Could not set up backend")
    try:
        if util.which(name) or backends[name].testAvailable():
            backend = backends[name]()
    except Exception:
        logging.exception("Failed to set up backend")

    play_sound = backend.play_sound
    stop_sound = backend.stop_sound
    is_playing = backend.is_playing
    resolve_sound = resolve_sound
    pause = backend.pause
    resume = backend.resume
    setvol = backend.set_volume
    position = backend.get_position
    fade_to = backend.fade_to
    readySound = backend.readySound
    preload = backend.preload


def stop_all_sounds():
    backend.stop_all_sounds()


def test(output=None):
    t = "Kaithemtest"
    play_sound("alert.ogg", output=output, handle=t)
    for i in range(100):
        if is_playing(t, refresh=True):
            return
        time.sleep(0.01)
    raise RuntimeError("Sound did not report as playing within 1000ms")


# Make fake module functions mapping to the bound methods.
# have to hardcode them here or else the linter might not be happy

play_sound = backend.play_sound
stop_sound = backend.stop_sound
is_playing = backend.is_playing
resolve_sound = resolve_sound
pause = backend.pause
resume = backend.resume
setvol = backend.set_volume
position = backend.get_position
fade_to = backend.fade_to
readySound = backend.readySound
preload = backend.preload

isStartDone = []

try:
    select_backend("mpv")
except Exception:
    logging.exception("Failed to select MPV backend.")



# coding: utf-8

from __future__ import annotations

import sys
import json
import io
import time
import threading

from typing import Any, Callable


class Spec(object):
    """
    This class wraps methods that create JSON-RPC 2.0 compatible string representations of
    request, response and error objects. All methods are class members, so you might never want to
    create an instance of this class, but rather use the methods directly:

    .. code-block:: python

        Spec.request("my_method", 18)  # the id is optional
        # => '{"jsonrpc":"2.0","method":"my_method","id": 18}'

        Spec.response(18, "some_result")
        # => '{"jsonrpc":"2.0","id":18,"result":"some_result"}'

        Spec.error(18, -32603)
        # => '{"jsonrpc":"2.0","id":18,"error":{"code":-32603,"message":"Internal error"}}'
    """

    @classmethod
    def check_id(cls, id: str | int | None, allow_empty: bool = False) -> None:
        """
        Value check for *id* entries. When *allow_empty* is *True*, *id* is allowed to be *None*.
        Raises a *TypeError* when *id* is neither an integer nor a string.
        """
        if (id is not None or not allow_empty) and not isinstance(id, (int, str)):
            raise TypeError(f"id must be an integer or string, got {id} ({type(id)})")

    @classmethod
    def check_method(cls, method: str) -> None:
        """
        Value check for *method* entries. Raises a *TypeError* when *method* is not a string.
        """
        if not isinstance(method, str):
            raise TypeError(f"method must be a string, got {method} ({type(method)})")

    @classmethod
    def check_code(cls, code: int) -> None:
        """
        Value check for *code* entries. Raises a *TypeError* when *code* is not an integer, or a
        *KeyError* when there is no :py:class:`RPCError` subclass registered for that *code*.
        """
        if not isinstance(code, int):
            raise TypeError(f"code must be an integer, got {code} ({type(code)})")

        if not get_error(code):
            raise ValueError(f"unknown code, got {code} ({type(code)})")

    @classmethod
    def request(
        cls,
        method: str,
        id: str | int | None = None,
        params: dict[str, Any] | None = None,
    ) -> str:
        """
        Creates the string representation of a request that calls *method* with optional *params*
        which are encoded by ``json.dumps``. When *id* is *None*, the request is considered a
        notification.
        """
        try:
            cls.check_method(method)
            cls.check_id(id, allow_empty=True)
        except Exception as e:
            raise RPCInvalidRequest(str(e))

        # start building the request string
        req = f'{{"jsonrpc":"2.0","method":"{method}"'

        # add the id when given
        if id is not None:
            # encode string ids
            if isinstance(id, str):
                id = json.dumps(id)
            req += f',"id":{id}'

        # add parameters when given
        if params is not None:
            try:
                req += f',"params":{json.dumps(params)}'
            except Exception as e:
                raise RPCParseError(str(e))

        # end the request string
        req += "}"

        return req

    @classmethod
    def response(cls, id: str | int | None, result: Any) -> str:
        """
        Creates the string representation of a respone that was triggered by a request with *id*.
        A *result* is required, even if it is *None*.
        """
        try:
            cls.check_id(id)
        except Exception as e:
            raise RPCInvalidRequest(str(e))

        # encode string ids
        if isinstance(id, str):
            id = json.dumps(id)

        # build the response string
        try:
            res = f'{{"jsonrpc":"2.0","id":{id},"result":{json.dumps(result)}}}'
        except Exception as e:
            raise RPCParseError(str(e))

        return res

    @classmethod
    def error(cls, id: str | int | None, code: int, data: dict | None = None) -> str:
        """
        Creates the string representation of an error that occured while processing a request with
        *id*. *code* must lead to a registered :py:class:`RPCError`. *data* might contain
        additional, detailed error information and is encoded by ``json.dumps`` when set.
        """
        try:
            cls.check_id(id)
            cls.check_code(code)
        except Exception as e:
            raise RPCInvalidRequest(str(e))

        # build the inner error data
        message = get_error(code).title
        err_data = f'{{"code":{code},"message":"{message}"'

        # insert data when given
        if data is not None:
            try:
                err_data += f',"data":{json.dumps(data)}}}'
            except Exception as e:
                raise RPCParseError(str(e))
        else:
            err_data += "}"

        # encode string ids
        if isinstance(id, str):
            id = json.dumps(id)

        # start building the error string
        err = f'{{"jsonrpc":"2.0","id":{id},"error":{err_data}}}'

        return err


class RPC(object):
    """
    The main class of *jsonrpyc*. Instances of this class wrap an input stream *stdin* and an output
    stream *stdout* in order to communicate with other services. A service is not even forced to be
    written in Python as long as it strictly implements the JSON-RPC 2.0 specification. RPC
    instances may wrap a *target* object. By means of a :py:class:`Watchdog` instance, incoming
    requests are routed to methods of this object whose result might be sent back as a response.
    The watchdog instance is created but not started yet, when *watch* is not *True*.
    Example implementation:

    *server.py*

    .. code-block:: python

        import jsonrpyc

        class MyTarget(object):

            def greet(self, name):
                return f"Hi, {name}!"

        jsonrpc.RPC(MyTarget())

    *client.py*

    .. code-block:: python

        import jsonrpyc
        from subprocess import Popen, PIPE

        p = Popen(["python", "server.py"], stdin=PIPE, stdout=PIPE)
        rpc = jsonrpyc.RPC(stdout=p.stdin, stdin=p.stdout)

        # non-blocking remote procedure call with callback and js-like signature
        def cb(err, res=None):
            if err:
                throw err
            print(f"callback got: {res}")

        rpc("greet", args=("John",), callback=cb)

        # cb is called asynchronously which prints
        # => "callback got: Hi, John!"

        # blocking remote procedure call with 0.1s polling
        print(rpc("greet", args=("John",), block=0.1))
        # => "Hi, John!"

        # shutdown the process
        p.stdin.close()
        p.stdout.close()
        p.terminate()
        p.wait()

    .. py:attribute:: target

        The wrapped target object. Might be *None* when no object is wrapped, e.g. for the *client*
        RPC instance.

    .. py:attribute:: stdin

        The input stream, re-opened with ``"rb"``.

    .. py:attribute:: stdout

        The output stream, re-opened with ``"wb"``.

    .. py:attribute:: watch

        The :py:class:`Watchdog` instance that optionally watches *stdin* and dispatches incoming
        requests.
    """

    EMPTY_RESULT = object()

    def __init__(
        self: RPC,
        target: Any | None = None,
        stdin: io.TextIOBase | None = None,
        stdout: io.TextIOBase | None = None,
        watch: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()

        # the wrapped target object
        self.target = target

        # open streams
        stdin = sys.stdin if stdin is None else stdin
        stdout = sys.stdout if stdout is None else stdout
        self.stdin = io.open(stdin.fileno(), "rb")
        self.stdout = io.open(stdout.fileno(), "wb")

        self.original_stdin = stdin

        # other attributes
        self._i = -1
        self._callbacks = {}
        self._results = {}

        # create and optionall start the watchdog
        kwargs["start"] = watch
        kwargs.setdefault("daemon", target is None)
        self.watchdog = Watchdog(self, **kwargs)

    def __del__(self: RPC) -> None:
        watchdog = getattr(self, "watchdog", None)
        if watchdog:
            watchdog.stop()

    def __call__(self: RPC, *args, **kwargs) -> None:
        """
        Shorthand for :py:meth:`call`.
        """
        return self.call(*args, **kwargs)

    def call(
        self: RPC,
        method: str,
        args: tuple[Any] = (),
        kwargs: dict | None = None,
        callback: Callable | None = None,
        block: int = 0,
        timeout: float = 0,
    ) -> None:
        """
        Performs an actual remote procedure call by writing a request representation (a string) to
        the output stream. The remote RPC instance uses *method* to route to the actual method to
        call with *args* and *kwargs*. When *callback* is set, it will be called with the result of
        the remote call. When *block* is larger than *0*, the calling thread is blocked until the
        result is received. In this case, *block* will be the poll interval, emulating synchronuous
        return value behavior. When both *callback* is *None* and *block* is *0* or smaller, the
        request is considered a notification and the remote RPC instance will not send a response.
        """
        starting_time = time.monotonic()

        # default kwargs
        if kwargs is None:
            kwargs = {}

        # check if the call is a notification
        is_notification = callback is None and block <= 0

        # create a new id for requests expecting a response
        id = None
        if not is_notification:
            self._i += 1
            id = self._i

        # register the callback
        if callback is not None:
            self._callbacks[id] = callback

        # store an empty result for the meantime
        if block > 0:
            self._results[id] = self.EMPTY_RESULT

        # create the request
        params = {"args": args, "kwargs": kwargs}
        req = Spec.request(method, id=id, params=params)
        self._write(req)

        # blocking return value behavior
        if block > 0:
            while True:
                if self._results[id] != self.EMPTY_RESULT:
                    result = self._results[id]
                    del self._results[id]
                    if isinstance(result, Exception):
                        raise result
                    return result

                if timeout:
                    elapsed = time.monotonic() - starting_time
                    if elapsed > timeout:
                        raise TimeoutError("RPC Request timed out")

                time.sleep(block)

    def _handle(self: RPC, line: str) -> None:
        """
        Handles an incoming *line* and dispatches the parsed object to the request, response, or
        error handlers.
        """
        try:
            obj = json.loads(line)
        except json.decoder.JSONDecodeError:
            if "BAD JSON" not in line:
                print(f"BAD JSON {line}")
            return

        # dispatch to the correct handler
        if "method" in obj:
            # request
            self._handle_request(obj)
        elif "error" not in obj:
            # response
            self._handle_response(obj)
        else:
            # error
            self._handle_error(obj)

    def _handle_request(self: RPC, req: dict[str, Any]) -> None:
        """
        Handles an incoming request *req*. When it containes an id, a response or error is sent
        back.
        """
        try:
            method = self._route(req["method"])
            result = method(*req["params"]["args"], **req["params"]["kwargs"])
            if "id" in req:
                res = Spec.response(req["id"], result)
                self._write(res)
        except Exception as e:
            if "id" in req:
                if isinstance(e, RPCError):
                    err = Spec.error(req["id"], e.code, e.data)
                else:
                    err = Spec.error(req["id"], -32603, str(e))
                self._write(err)

    def _handle_response(self: RPC, res: dict[str, Any]) -> None:
        """
        Handles an incoming successful response *res*. Blocking calls are resolved and registered
        callbacks are invoked with the first error argument being set to *None*.
        """
        # set the result
        if res["id"] in self._results:
            self._results[res["id"]] = res["result"]

        # lookup and invoke the callback
        if res["id"] in self._callbacks:
            callback = self._callbacks[res["id"]]
            del self._callbacks[res["id"]]
            callback(None, res["result"])

    def _handle_error(self: RPC, res: dict[str, Any]) -> None:
        """
        Handles an incoming failed response *res*. Blocking calls throw an exception and
        registered callbacks are invoked with an exception and the second result argument set to
        *None*.
        """
        # extract the error and create an actual error instance to raise
        err = res["error"]
        error = get_error(err["code"])(err.get("data", err["message"]))

        # set the error
        if res["id"] in self._results:
            self._results[res["id"]] = error

        # lookup and invoke the callback
        if res["id"] in self._callbacks:
            callback = self._callbacks[res["id"]]
            del self._callbacks[res["id"]]
            callback(error, None)

    def _route(self: RPC, method: str) -> Any:
        """
        Returnes the method of the wrapped target object to be called when *method* is requested.
        Example:

        .. code-block:: python

            MyClassB(object):
                def foo(self):
                    return 123

            MyClassA(object):
                def __init__(self):
                    self.b = MyClassB()

                def bar(self):
                    return "test"

            rpc = RPC(MyClassA())

            rpc._route("bar")
            # => <bound method MyClassA.bar ...>

            rpc._route("b.foo")
            # => <bound method MyClassB.foo ...>
        """
        # recursively traverse target attributes
        obj = self.target
        for part in method.split("."):
            if not hasattr(obj, part):
                break
            obj = getattr(obj, part)
        else:
            return obj

        raise RPCMethodNotFound(data=method)

    def _write(self: RPC, s: str) -> None:
        """
        Writes a string *s* to the output stream.
        """
        self.stdout.write(bytearray(s + "\n", "utf-8"))
        self.stdout.flush()


class Watchdog(threading.Thread):
    """
    This class represents a thread that watches the input stream of an :py:class:`RPC` instance for
    incoming content and dispatches requests to it.

    .. py:attribute:: rpc

        The :py:class:`RPC` instance.

    .. py:attribute:: name

        The thread's name.

    .. py:attribute:: interval

        The polling interval of the run loop.

    .. py:attribute:: daemon

        The thread's daemon flag.
    """

    def __init__(
        self: Watchdog,
        rpc: RPC,
        name: str = "watchdog",
        interval: float = 0.1,
        daemon: bool = False,
        start: bool = True,
    ) -> None:
        super().__init__()

        # store attributes
        self.rpc = rpc
        self.name = name
        self.interval = interval
        self.daemon = daemon

        # register a stop event
        self._stop = threading.Event()

        if start:
            self.start()

    def start(self: Watchdog) -> None:
        """
        Starts with thread's activity.
        """
        super().start()

    def stop(self: Watchdog) -> None:
        """
        Stops with thread's activity.
        """
        self._stop.set()

    def run(self: Watchdog) -> None:
        # reset the stop event
        self._stop.clear()

        # stop here when stdin is not set or closed
        if not self.rpc.stdin or self.rpc.stdin.closed:
            return

        # read new incoming lines
        last_pos = 0
        while not self._stop.is_set():
            lines = None

            # stop when stdin is closed
            if self.rpc.stdin.closed:
                break

            if self.rpc.original_stdin and self.rpc.original_stdin.closed:
                break

            # read from stdin depending on whether it is a tty or not
            if self.rpc.stdin.isatty():
                cur_pos = self.rpc.stdin.tell()
                if cur_pos != last_pos:
                    self.rpc.stdin.seek(last_pos)
                    lines = self.rpc.stdin.readlines()
                    last_pos = self.rpc.stdin.tell()
                    self.rpc.stdin.seek(cur_pos)
            else:
                try:
                    lines = [self.rpc.stdin.readline()]
                except (IOError, ValueError):
                    # prevent residual race conditions occurring when stdin is closed externally
                    pass

            # handle new lines if any
            if lines:
                for line in lines:
                    line = line.decode("utf-8").strip()
                    if line:
                        self.rpc._handle(line)
            else:
                self._stop.wait(self.interval)


class RPCError(Exception):
    """
    Base class for RPC errors.

    .. py:attribute:: message

        The message of this error, i.e., ``"<title> (<code>)[, data: <data>]"``.

    .. py:attribute:: data

        Additional data of this error. Setting the data attribute will also change the message
        attribute.
    """

    def __init__(self: RPCError, data: str | None = None) -> None:
        # build the error message
        message = f"{self.title} ({self.code})"
        if data is not None:
            message += f", data: {data}"
        self.message = message

        super().__init__(message)

        self.data = data

    def __str__(self):
        return self.message


error_map_distinct = {}
error_map_range = {}


def is_range(code: Any) -> bool:
    return (
        isinstance(code, tuple)
        and len(code) == 2
        and all(isinstance(i, int) for i in code)
        and code[0] < code[1]
    )


def register_error(cls: type) -> type:
    """
    Decorator that registers a new RPC error derived from :py:class:`RPCError`. The purpose of
    error registration is to have a mapping of error codes/code ranges to error classes for faster
    lookups during error creation.

    .. code-block:: python

        @register_error
        class MyCustomRPCError(RPCError):
            code = ...
            title = "My custom error"
    """
    # it would be much cleaner to add a meta class to RPCError as a registry for codes
    # but in CPython 2 exceptions aren't types, so simply provide a registry mechanism here
    if not issubclass(cls, RPCError):
        raise TypeError(f"'{cls}' is not a subclass of RPCError")

    code = cls.code

    if isinstance(code, int):
        error_map = error_map_distinct
    elif is_range(code):
        error_map = error_map_range
    else:
        raise TypeError(f"invalid RPC error code {code}")

    if code in error_map:
        raise AttributeError(f"duplicate RPC error code {code}")

    error_map[code] = cls

    return cls


def get_error(code: int) -> type | None:
    """
    Returns the RPC error class that was previously registered to *code*. *None* is returned when no
    class could be found.
    """
    if code in error_map_distinct:
        return error_map_distinct[code]

    for (lower, upper), cls in error_map_range.items():
        if lower <= code <= upper:
            return cls

    return None


@register_error
class RPCParseError(RPCError):
    code = -32700
    title = "Parse error"


@register_error
class RPCInvalidRequest(RPCError):
    code = -32600
    title = "Invalid Request"


@register_error
class RPCMethodNotFound(RPCError):
    code = -32601
    title = "Method not found"


@register_error
class RPCInvalidParams(RPCError):
    code = -32602
    title = "Invalid params"


@register_error
class RPCInternalError(RPCError):
    code = -32603
    title = "Internal error"


@register_error
class RPCServerError(RPCError):
    code = (-32099, -32000)
    title = "Server error"



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

from __future__ import annotations

import threading
import time
import logging
import weakref
import traceback
import os
import sys
import base64
import math
import gi

gi.require_version("Gst", "1.0")
gi.require_version("GstBase", "1.0")
gi.require_version("Gtk", "3.0")


from gi.repository import Gst  # noqa


def doNow(f):
    f()


lock = threading.RLock()


Gst.init(None)

jackChannels = {}

stopflag = [0]

# Overridden later
print = print


def dummy_func(*a, **k):
    "RPC replaces this with the thing it calls"


rpc = [dummy_func]


def call_rpc_if_exists(*a, **k):
    if rpc[0]:
        rpc[0](*a, **k)


# https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python


def check_pid(pid: int):
    """Check For the existence of a unix pid."""
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


pipes = weakref.WeakValueDictionary()

log = logging.getLogger("IceFlow_gst")


# This is NixOS compatibility stuff, we could be running as an output from setup.py
# Or we could be running directly with python3 file.py
try:
    from . import jsonrpyc
except ImportError:
    import jsonrpyc


class PresenceDetectorRegion:
    def __init__(self):
        # This is a first order filter(Time domain blur) of the entire image

        self.state = None
        self.last = None

    def poll(self, val):
        #
        from PIL import ImageFilter
        from PIL import ImageChops

        try:
            import scipy.ndimage

            ndim = 1
        except Exception:
            ndim = 0

        import numpy as np

        x = val

        self.last = x if x else self.last

        rval = 0
        if self.state and self.last:
            diff = ImageChops.difference(self.state, self.last)
            # This is an erosion operation to prioritize multipixel stuff
            # over single pixel noise
            d = diff.convert("F")

            if ndim:
                # This is like 4 times faster.
                d = scipy.ndimage.grey_erosion(d, (3, 3))
            else:
                d = d.filter(ImageFilter.MinFilter(3))
            d = np.array(d)

            # Ignore everythong below the threshold, that gets rid of a lot of our noise
            m = np.mean(d) * 1.5 + 4
            d = np.fmax(d - m, 0)

            x = np.mean(d * d)
            if x == 0:
                rval = 0

            rval = math.sqrt(x) / 2.5

        self.state = self.last
        return rval


class PresenceDetector:
    def __init__(self, capture, regions=None):
        self.masks = regions
        self.regions = {}
        self.entireImage = PresenceDetectorRegion()
        self.capture = capture

        if regions is not None:
            for i in regions:
                self.regions[i] = PresenceDetectorRegion()

    def poll(self):
        r = {}

        x = self.capture.pull()
        if x is None:
            return None

        w = x.width
        h = x.height

        if self.masks is None:
            return self.entireImage.poll(x)
        else:
            r[""] = self.entireImage.poll(x)

        for i in self.masks:
            m = self.masks[i]
            # Crop region is specified as a fraction, convert to pixels and points instead of fraction y,x,w,h
            i2 = x.crop(
                (
                    int(m[0] * w),
                    int(m[1] * h),
                    int(m[0] * w) + int(m[2] * w),
                    int(m[1] * h) + int(m[3] * h),
                )
            )
            r[i] = self.regions[i].poll(i2)
        return r


class PILCapture:
    def __init__(self, appsink):
        from PIL import Image

        self.img = Image
        self.appsink = appsink

    def pull_to_file(self, f, timeout=0.1):
        x = self.pull(timeout, True)
        if not x:
            return None
        x.save(f)
        return 1

    def pull(self, timeout=0.1, force_latest=False):
        sample = self.appsink.emit("try-pull-sample", timeout * 10**9)

        if force_latest:
            # Try another pull but only wait 1ms.
            # This is in case there is another queued up frame, such as if we have buffer elements or something
            # before this
            sample2 = self.appsink.emit("try-pull-sample", 1000000)
            c = 10
            while sample2:
                if c < 1:
                    break
                c -= 1
                sample2 = self.appsink.emit("try-pull-sample", 1000000)

            sample = sample2 or sample

        if not sample:
            return None

        buf = sample.get_buffer()
        caps = sample.get_caps()
        h = caps.get_structure(0).get_value("height")
        w = caps.get_structure(0).get_value("width")

        return self.img.frombytes("RGB", (w, h), buf.extract_dup(0, buf.get_size()))

    def pull_raw(self):
        "Pull a tuple consisting of raw RGB bytes, then the height and width with which to decode them."
        sample = self.appsink.emit("try-pull-sample", 0.1 * 10**9)
        if not sample:
            return None

        buf = sample.get_buffer()
        caps = sample.get_caps()
        h = caps.get_structure(0).get_value("height")
        w = caps.get_structure(0).get_value("width")

        return (buf.extract_dup(0, buf.get_size()), w, h)


class PILSource:
    def __init__(self, appsrc, greyscale=False):
        self.appsrc = appsrc
        self.greyscale = greyscale

    def push(self, img):
        img = img.tobytes("raw", "L" if self.greyscale else "rgb")
        img = Gst.Buffer.new_wrapped(img)
        self.appsrc.emit("push-buffer", img)


class AppSource:
    def __init__(self, appsrc):
        self.appsrc = appsrc

    def push(self, b):
        b = Gst.Buffer.new_wrapped(b)
        self.appsrc.emit("push-buffer", b)


class AppSink:
    # Used to pull the raw bytes buffer data.
    def __init__(self, appsink):
        self.appsink = appsink

    def pull(self, timeout=0.1):
        sample = self.appsink.emit("try-pull-sample", timeout * 10**9)
        if not sample:
            return None

        buf = sample.get_buffer()
        sample.get_caps()

        return buf.extract_dup(0, buf.get_size())


def link(a, b):
    unref = False
    try:
        if not a or not b:
            raise ValueError("Cannot link None")
        if isinstance(a, Gst.Pad):
            if not isinstance(b, Gst.Pad):
                b = b.get_static_pad("sink")
                unref = True
                if not b:
                    raise RuntimeError("B has no pad named sink and A is a pad")
            if not a.link(b) == Gst.PadLinkReturn.OK:
                raise RuntimeError("Could not link: " + str(a) + str(b))

        else:
            x = a.link(b)

            if not x:
                raise RuntimeError(
                    "Could not link" + str(a) + str(b) + " reason " + str(x)
                )
    finally:
        if unref:
            pass  # b.unref()


def elementInfo(e):
    r = Gst.Registry.get()  # noqa


elementsByShortId = weakref.WeakValueDictionary()


def Element(n, name=None):
    e = Gst.ElementFactory.make(n, None)
    elementsByShortId[id(e)] = e
    if e:
        return e
    else:
        raise ValueError("No such element exists: " + n)


loop = None

autorunMainloop = True

mainContext = None


def does_element_exist(n):
    n = Gst.ElementFactory.make(n)
    if n:
        pass  # n.unref()
    return True


def wrfunc(f, fail_return=None):
    def f2(*a, **k):
        try:
            return f()(*a, **k)
        except Exception:
            print(traceback.format_exc())
            return fail_return

    return f2


def makeWeakrefPoller(selfref, exitSignal):
    def pollerf():
        t = 0
        seq_num = -1
        while selfref():
            self = selfref()

            self.thread_started = True
            if not self.shouldRunThread:
                exitSignal.append(True)

                # We can't allow returns to happen till the pipeline is null.  That could cause a segfault
                # on garbage collection.  So we choose a memory leak instead.
                try:
                    with self.seeklock:
                        self.pipeline.set_state(Gst.State.NULL)
                except Exception:
                    pass
                self._waitForState(Gst.State.NULL, 3600000)
                return

            if self.shouldRunThread:
                if t < time.monotonic() - 3:
                    t = time.monotonic()
                    try:
                        # If we can't get the lock, don't block the message thread
                        if self.lock.acquire(timeout=1):
                            try:
                                self.loopCallback()
                            finally:
                                self.lock.release()

                        state = self.pipeline.get_state(1000000000)[1]
                        if not state == Gst.State.NULL:
                            self.wasEverRunning = True

                        # Set the flag if anything ever drives us into the null state
                        if self.wasEverRunning and state == Gst.State.NULL:
                            self.shouldRunThread = False
                            exitSignal.append(True)
                            return
                    except Exception:
                        # Todo actually handle some errors?
                        exitSignal.append(True)
                        # After pipeline deleted, we clean up
                        if hasattr(self, "pipeline") and hasattr(self, "bus"):
                            try:
                                with self.seeklock:
                                    self.pipeline.set_state(Gst.State.NULL)
                            except Exception:
                                pass
                            self._waitForState(Gst.State.NULL, 3600000)

                            raise
                        else:
                            return

            # with self.lock:
            msg = self.bus.timed_pop(500 * 1000 * 1000)
            if msg:
                try:
                    if msg.type == Gst.MessageType.ERROR:
                        self.on_error(self.bus, msg, None)

                    elif msg.type == Gst.MessageType.EOS:
                        if msg.seqnum != seq_num:
                            self._on_eos(self.bus, msg, None)

                    elif msg.type == Gst.MessageType.SEGMENT_DONE:
                        if msg.seqnum != seq_num:
                            self._on_segment_done()

                    self._on_message(self.bus, msg, None)

                    seq_num = msg.seqnum
                except Exception:
                    sys.stderr.write(traceback.format_exc())
                finally:
                    pass
            else:
                # Too quiet in here and the seeklock is taken, assume the seek was jammed
                # by a move to the pause state
                if self.seeklock.acquire(timeout=0.25):
                    self.seeklock.release()
                else:
                    self.pipeline.set_state(Gst.State.PLAYING)

            del self

            # time.sleep(1)

    return pollerf


def getCaps(e):
    try:
        return e.caps
    except Exception:
        return "UNKNOWN"
    e.getSinks()[0].getNegotiatedCaps()


def linkClosureMaker(
    self, src, dest, connect_when_available, eid, deleteAfterUse=False
):
    "This has t be outside, it can't leak a ref to self's strong reference into the closure or it may leak memory"

    def linkFunction(element, pad, dummy):
        s = pad.query_caps(None).to_string()
        if isinstance(connect_when_available, str):
            if connect_when_available not in s:
                return

        if eid in self().waitingCallbacks:
            link(element, dest)
        if deleteAfterUse:
            del self().waitingCallbacks[eid]

    return linkFunction


class GStreamerPipeline:
    """Semi-immutable pipeline that presents a nice subclassable GST pipeline You can only add stuff to it."""

    def __init__(self, name=None, realtime=None, system_time=False):
        self.lock = threading.RLock()

        self.seeklock = self.lock

        self.pilcaptures = []

        self.appsink = None
        # We use this for detecting motion.
        # We have to use this hack because gstreamer's detection is... not great.
        self._pilmotiondetectorcapture = None
        self._pilmotiondetector = None

        self.exiting = False

        self.uuid = time.time()
        name = name or "Pipeline" + str(time.monotonic())
        self.realtime = realtime
        self._stopped = True
        self.sync_stop = False
        self.pipeline = Gst.Pipeline()
        self.thread_started = False
        self.weakrefs = weakref.WeakValueDictionary()

        self.proxyToElement = weakref.WeakValueDictionary()

        # This WeakValueDictionary is mostly for testing purposes
        pipes[id(self)] = self

        # Thread puts something in this so we know we exited
        self.exitSignal = []

        self.weakrefs[str(self.pipeline)] = self.pipeline
        if not self.pipeline:
            raise RuntimeError("Could not create pipeline")

        self.bus = self.pipeline.get_bus()
        self._stopped = False

        self.weakrefs[str(self.bus)] = self.bus

        self.hasSignalWatch = 0

        # Run in our own mainContext
        # mainContext.push_thread_default()
        try:
            self.bus.add_signal_watch()
        finally:
            pass  # mainContext.pop_thread_default()

        self.hasSignalWatch = 1
        # 1 is dummy user data, because some have reported segfaults if it is missing
        # Note that we keep strong refs to the functions, so they don't go away when we unregister,
        # Leading to a segfault in libffi because of a race condition
        self._on_message_wr = wrfunc(weakref.WeakMethod(self._on_message))
        self.pgbcobj = self.bus.connect("message", self._on_message_wr, 1)

        self._on_eos_wr = wrfunc(weakref.WeakMethod(self._on_eos))
        self.pgbcobj2 = self.bus.connect("message::eos", self._on_eos_wr, 1)

        self._onerror_wr = wrfunc(weakref.WeakMethod(self.on_error))
        self.pgbcobj3 = self.bus.connect("message::error", self._onerror_wr, 1)

        self.name = name

        self.elements = []
        self.sidechainElements = []
        self.namedElements = {}

        self.elementTypesById = {}

        # Just a place to store refs
        self.waitingCallbacks = {}

        self.running = False
        self.shouldRunThread = True
        self.wasEverRunning = True

        self.knownThreads = {}
        self.startTime = 0

        def dummy(*a, **k):
            pass

        if realtime:
            self._syncmessage = wrfunc(
                weakref.WeakMethod(self.syncMessage), fail_return=Gst.BusSyncReply.PASS
            )
            self.bus.set_sync_handler(self._syncmessage, 0, dummy)
        self.pollthread = None

        self.lastElementType = None

        # If true, keep the pipeline running at the same rate as system time
        self.system_time = system_time
        self.targetRate = 1.0
        self.pipelineRate = 1.0

    def sendEOS(self):
        self.pipeline.send_event(Gst.Event.new_eos())

    def loopCallback(self):
        if self._pilmotiondetector:
            x = self._pilmotiondetector.poll()
            if x is None:
                return
            call_rpc_if_exists("on_presence_value", [x])

    def add_presence_detector(self, resolution, connect_to_output=None, regions=None):
        if self._pilmotiondetector:
            raise RuntimeError("Already have one of these")

        self._pilmotiondetectorcapture = self.add_pil_capture(
            resolution, connect_to_output, method=0
        )
        self._pilmotiondetector = PresenceDetector(
            self._pilmotiondetectorcapture, regions
        )

    def seek(
        self,
        t=None,
        rate=None,
        _raw=False,
        _offset=0.008,
        flush=True,
        segment=False,
        sync=False,
        skip=False,
    ):
        "Seek the pipeline to a position in seconds, set the playback rate, or both"
        with self.lock:
            if self.exiting:
                return
            if not self.running:
                return

            if rate is None:
                rate = self.targetRate

            if not _raw:
                # Set "effective start time" so that the system clock sync keeps working.
                if t is not None:
                    t = max(t, 0)
                    self.startTime = time.monotonic() - t
                self.targetRate = rate
                self.pipelineRate = rate

            flags = Gst.SeekFlags.NONE

            if flush:
                flags |= Gst.SeekFlags.FLUSH

            if segment:
                flags |= Gst.SeekFlags.SEGMENT

            if skip:
                # Use skip to speed up, but segment mode
                flags |= Gst.SeekFlags.SKIP

            if (not flush) and self.pipeline.get_state(1000_000_000)[
                1
            ] == Gst.State.PAUSED:
                raise RuntimeError(
                    "Cannot do non-flushing seek in paused state as this may deadlock"
                )

            # Big issue here.  Sometimes, i think when you stop playback during seek,
            # It blocks forever.  We have to do this in a background thread so we don't block everything
            # Else.   We just have to trust gstreamer to be thread safe, and to eventually unblock when the pipeline is null.
            e = threading.Event()

            def f():
                # Seek is especiallly problematic, we must isolate it from state changes
                with self.seeklock:
                    if (not flush) and self.pipeline.get_state(1000_000_000)[
                        1
                    ] == Gst.State.PAUSED:
                        raise RuntimeError(
                            "Cannot do non-flushing seek in paused as this may deadlock.  State has changed while request inflight, request cancelled."
                        )

                    self.pipeline.seek(
                        rate,
                        Gst.Format.TIME,
                        flags,
                        Gst.SeekType.NONE if t is None else Gst.SeekType.SET,
                        max(((t or 0) + _offset) * 10**9, 0),
                        Gst.SeekType.NONE,
                        0,
                    )
                    e.set()

            if sync:
                f()
            else:
                doNow(f)
            # If it got blocked because a race condition that made it paused, try to unblock
            if not e.wait(0.5):
                if not flush:
                    self.pipeline.set_state(Gst.State.PLAYING)

    def getPosition(self):
        "Returns stream position in seconds"
        with self.lock:
            ret, current = self.pipeline.query_position(Gst.Format.TIME)
            if not ret:
                raise RuntimeError(ret)
            if current < 0:
                raise RuntimeError("Nonsense position: " + str(current))
            if current == Gst.CLOCK_TIME_NONE:
                raise RuntimeError("gst.CLOCK_TIME_NONE")
            return current / 10**9

    @staticmethod
    def setcurrent_threadPriority(x, y):
        raise RuntimeError("Must override this to use realtime priority")

    def syncMessage(self, *arguments):
        "Synchronous message, so we can enable realtime priority on individual threads."
        # Stop the poorly performing sync messages after a while.
        # Wait till we have at least one thread though.

        try:
            if self.knownThreads and time.monotonic() - self.startTime > 3:
                # This can't use the lock, we don't know what thread it might be called in.
                def noSyncHandler():
                    with self.lock:
                        if hasattr(self, "bus"):
                            self.bus.set_sync_handler(None, 0, None)

                doNow(noSyncHandler)
            if threading.current_thread().ident not in self.knownThreads:
                self.knownThreads[threading.current_thread().ident] = True
                if self.realtime:
                    try:
                        self.setcurrent_threadPriority(1, self.realtime)
                    except Exception:
                        log.exception("Error setting realtime priority")
            return Gst.BusSyncReply.PASS
        except Exception:
            return Gst.BusSyncReply.PASS
            print(traceback.format_exc())

    def makeElement(self, n, name=None):
        with self.lock:
            e = Element(n, name)
            self.elementTypesById[id(e)] = n
            self.pipeline.add(e)
            return e

    # Low level wrapper just for filtering out args we don't care about
    def _on_eos(self, *a, **k):
        with self.lock:
            self.on_eos()

    def on_eos(self):
        self.on_stream_finished()
        call_rpc_if_exists("on_stream_finished", [])

    def on_stream_finished(self):
        pass

    def __del__(self):
        self.running = False
        t = time.monotonic()

        # Give it some time, in case it really was started
        if not self.thread_started:
            time.sleep(0.01)
            time.sleep(0.01)
            time.sleep(0.01)
            time.sleep(0.01)

        if self.thread_started:
            while not self.exitSignal:
                time.sleep(0.1)
                if time.monotonic() - t > 10:
                    raise RuntimeError("Timeout")

        if not self._stopped:
            self.stop()

    def _on_message(self, bus, message, userdata):
        s = message.get_structure()
        if s:
            self.on_message(message.src, s.get_name(), s)

        return True

    # def appsinkhandler(self,appsink, user_data):
    #     sample = appsink.emit("pull-sample")
    #     gst_buffer = sample.get_buffer()
    #     (ret, buffer_map) = gst_buffer.map(Gst.MapFlags.READ)
    #     call_rpc_if_exists("_on_appsink_data", [str(user_data), base64.b64encode(buffer_map.data).decode()])
    #     return Gst.FlowReturn.OK

    def on_message(self, src, name, s):
        if s.get_name() == "level":
            rms = sum([i for i in s["rms"]]) / len(s["rms"])
            decay = sum([i for i in s["decay"]]) / len(s["decay"])
            call_rpc_if_exists("on_level_message", [str(src), rms, decay])

        elif s.get_name() == "motion":
            if s.has_field("motion_begin"):
                call_rpc_if_exists("on_motion_begin", [])
            if s.has_field("motion_finished"):
                call_rpc_if_exists("on_motion_end", [])

        elif s.get_name() == "GstVideoAnalyse":
            call_rpc_if_exists(
                "on_video_analyze",
                [
                    {
                        "luma-average": s.get_double("luma-average")[1],
                        "luma-variance": s.get_double("luma-variance")[1],
                    }
                ],
            )

        elif s.get_name() == "barcode":
            call_rpc_if_exists(
                "on_barcode",
                [s.get_string("type"), s.get_string("symbol"), s.get_int("quality")[1]],
            )

        elif s.get_name() == "GstMultiFileSink":
            call_rpc_if_exists("on_multi_file_sink", [""])

        elif s.get_name() == "pocketsphynx":
            if s.get_value("hypothesis"):
                call_rpc_if_exists(
                    "onSTTMessage", [str(src), s.get_value("hypothesis")]
                )
            if s.get_value("final"):
                call_rpc_if_exists(
                    "onSTTMessageFinal", [str(src), s.get_value("final")]
                )

    def on_error(self, bus, msg, userdata):
        with self.lock:
            logging.debug("Error {}: {}, {}".format(msg.src.name, *msg.parse_error()))

    def _on_segment_done(self, *a):
        with self.lock:
            self.on_segment_done()
            return True

    def on_segment_done(self):
        # Called when a segment finishes playback, but NOT whem a segment ends because you did a seek to a new segment,
        # As that is usually not what you want when doing seamless loops.
        call_rpc_if_exists("on_segment_done", [])

    def _waitForState(self, s, timeout=10):
        t = time.monotonic()
        i = 0.01
        while not self.pipeline.get_state(1000_000_000)[1] == s:
            if time.monotonic() - t > timeout:
                raise RuntimeError(
                    "Timeout, pipeline still in: ",
                    self.pipeline.get_state(1000_000_000)[1],
                )
            time.sleep(min(i, 0.1))
            i *= 2

    def exitSegmentMode(self):
        with self.lock:
            self.seek()

    def restart(self, segment=False):
        # Todo why do we even need the global lock here?
        def f():
            with self.lock:
                if not self.pipeline.get_state(1000_000_000)[1] == Gst.State.NULL:
                    with self.seeklock:
                        self.pipeline.set_state(Gst.State.NULL)
                    self._waitForState(Gst.State.NULL)
                self.start(segment=segment)

        doNow(f)

    def start(self, effectiveStartTime=None, timeout=10, segment=False):
        "effectiveStartTime is used to keep multiple players synced when used with system_time"
        with self.lock:
            if self.exiting:
                return

            x = effectiveStartTime or time.time()
            timeAgo = time.time() - x
            # Convert to monotonic time that the nternal APIs use
            self.startTime = time.monotonic() - timeAgo

            # Go straight to playing, no need to locally do paused if we aren't using that feature
            if self.system_time or effectiveStartTime or segment:
                if not self.pipeline.get_state(1000_000_000)[1] == (
                    Gst.State.PAUSED,
                    Gst.State.PLAYING,
                ):
                    with self.seeklock:
                        self.pipeline.set_state(Gst.State.PAUSED)
                    self._waitForState(Gst.State.PAUSED)

            # Seek to where we should be, if we had actually
            # Started when we should have. We want to get everything set up in the pause state
            # First so we have the right "effective" start time.

            # We accept cutting off a few 100 milliseconds if it means
            # staying synced.
            if self.system_time:
                self.seek(time.monotonic() - self.startTime)

            elif segment:
                self.seek(0, segment=True, flush=True)

            with self.seeklock:
                self.pipeline.set_state(Gst.State.PLAYING)

            self._waitForState(Gst.State.PLAYING, timeout)

            self.running = True

            for i in range(500):
                try:
                    # Test that we can actually read the clock
                    self.getPosition()
                    break
                except Exception:
                    if i > 150:
                        raise RuntimeError("Clock still not valid")
                    time.sleep(0.1)

            # Don't start the thread until we have a valid clock
            self.maybeStartPoller()

    def play(self, segment=False):
        with self.lock:
            if self.exiting:
                return
            if not self.running:
                raise RuntimeError("Pipeline is not paused, or running, call start()")
            if self.pipeline.get_state(1000_000_000)[1] not in (
                Gst.State.PLAYING,
                Gst.State.PAUSED,
                Gst.State.READY,
            ):
                raise RuntimeError("Pipeline is not paused, or running, call start()")

            # Hopefully this willl raise an error if the clock is invalid for some reason,
            # Instead of potentially causing a segfault, if that was the problem
            self.getPosition()
            with self.seeklock:
                self.pipeline.set_state(Gst.State.PLAYING)
            self._waitForState(Gst.State.PLAYING)

            if segment:
                self.seek(segment=True, flush=False)

    def pause(self):
        "Not that we can start directly into paused without playing first, to preload stuff"
        with self.lock:
            if self.exiting:
                return
            with self.seeklock:
                self.pipeline.set_state(Gst.State.PAUSED)
            self._waitForState(Gst.State.PAUSED)
            self.getPosition()
            self.running = True
            self.maybeStartPoller()

    def maybeStartPoller(self, join=False):
        if not self.pollthread:
            self.pollthread = threading.Thread(
                target=makeWeakrefPoller(weakref.ref(self), self.exitSignal),
                daemon=True,
                name="nostartstoplog.GSTPoller",
            )
            self.pollthread.daemon = True
            self.pollthread.start()
            if join:
                self.pollthread.join()

    def isStoppedOrExiting(self):
        if self._stopped:
            return True
        if self.exiting:
            return True

    def stop(self):
        try:
            # Actually stop as soon as we can

            # Here again we must trust GST thread safety, or else we a blocked up pipeline
            # Could hold the lock and we could never stop it.
            self.shouldRunThread = False

            if not self.exiting:
                if hasattr(self, "pipeline"):
                    with self.seeklock:
                        # This was causing segfaults for some reasons
                        if not (
                            self.pipeline.get_state(1000_000_000)[1] == Gst.State.NULL
                        ):
                            # The set state line seemed to be a problem, better set the exiting
                            # flag early so we don't do it more than needed and hang?
                            self.exiting = True
                            self.pipeline.set_state(Gst.State.NULL)

            with self.lock:
                if self._stopped:
                    return

                self.exiting = True
                if hasattr(self, "bus"):
                    self.bus.set_sync_handler(None, 0, None)
                    if self.hasSignalWatch:
                        self.bus.remove_signal_watch()
                        self.hasSignalWatch = False

                # Now we're going to do the cleanup stuff
                # In the background, because it involves a lot of waiting.
                # This might fail, if it never even started, but we just kinda ignore that.p
                self.running = False
                self.shouldRunThread = False
                t = time.monotonic()
                time.sleep(0.01)

                if not self.thread_started:
                    time.sleep(0.01)
                    time.sleep(0.01)
                    time.sleep(0.01)
                    time.sleep(0.01)

                # On account of the race condition, it is possible that the thread actually never did start yet
                # So we have to ignore the exit flag stuff.

                # It shouldn't really be critical, most likely the thread can stop on it's own time anyway,
                # because it doesn't do anything without getting the lock.
                if self.thread_started:
                    while not self.exitSignal:
                        time.sleep(0.1)
                        if time.monotonic() - t > 10:
                            break

                with self.lock:
                    if self._stopped:
                        return
                    try:
                        self._waitForState(Gst.State.NULL, 1)
                    except Exception:
                        with self.seeklock:
                            self.pipeline.set_state(Gst.State.NULL)
                        self._waitForState(Gst.State.NULL, 1)

                    # This stuff happens in the NULL state, because we prefer not to mess with stuff while it's
                    # Running
                    try:
                        self.bus.disconnect(self.pgbcobj)
                        del self.pgbcobj
                        self.bus.disconnect(self.pgbcobj2)
                        del self.pgbcobj2
                        self.bus.disconnect(self.pgbcobj3)
                        del self.pgbcobj3
                    except Exception:
                        print(traceback.format_exc())

                    def f():
                        with lock:
                            try:
                                del jackChannels[self.uuid]
                            except KeyError:
                                pass

                    doNow(f)

                    self._stopped = True
        finally:
            stopflag[0] = 1

    def add_pil_capture(
        self, resolution=None, connect_to_output=None, buffer=1, method=1
    ):
        "Return a video capture object.  Now that we use BG threads this is just used to save snapshots to file"
        if resolution:
            self.add_element("videoscale", method=method)
            self.add_element(
                "capsfilter",
                caps="video/x-raw,width="
                + str(resolution[0])
                + ",height="
                + str(resolution[0]),
            )
        self.add_element("videoconvert", connect_to_output=connect_to_output)
        self.add_element("capsfilter", caps="video/x-raw,format=RGB")

        appsink = self.add_element("appsink", drop=True, sync=False, max_buffers=buffer)

        p = PILCapture(appsink)
        elementsByShortId[id(p)] = p
        self.pilcaptures.append(p)
        return p

    def addRemotePILCapture(self, *a, **k):
        return id(self.add_pil_capture(*a, **k))

    def addPILSource(self, resolution, buffer=1, greyscale=False):
        "Return a video source object that we can use to put PIL buffers into the stream"

        appsrc = self.add_element(
            "appsrc",
            caps="video/x-raw,width="
            + str(resolution[0])
            + ",height="
            + str(resolution[0])
            + ", format="
            + "GREy8"
            if greyscale
            else "RGB",
            connect_to_output=False,
        )
        self.add_element("videoconvert")
        self.add_element("videoscale")

        # Start with a blck image to make things prerooll
        if greyscale:
            appsrc.emit(
                "push-buffer",
                Gst.Buffer.new_wrapped(bytes(resolution[0] * resolution[1])),
            )
        else:
            appsrc.emit(
                "push-buffer",
                Gst.Buffer.new_wrapped(bytes(resolution[0] * resolution[1] * 3)),
            )

        return PILSource(appsrc, greyscale)

    def pull_buffer(self, element, timeout=0.1):
        if not self.appsink:
            raise RuntimeError("No appsink")

        if isinstance(element, int):
            element = elementsByShortId[element]

        sample = self.appsink.emit("try-pull-sample", timeout * 10**9)
        if not sample:
            return None

        buf = sample.get_buffer()
        sample.get_caps()

        return base64.b64encode(buf.extract_dup(0, buf.get_size()))

    def pull_to_file(self, element, fn):
        if isinstance(element, int):
            element = elementsByShortId[element]

        return element.pull_to_file(fn)

    def add_element(
        self,
        t,
        name=None,
        connect_when_available=False,
        connectWhenAvailable=False,
        connect_to_output=None,
        connectToOutput=None,
        sidechain=False,
        **kwargs,
    ):
        # TODO: remove Legacy hack eventually
        connect_to_output = connect_to_output or connectToOutput
        connect_when_available = connect_when_available or connectWhenAvailable
        with self.lock:
            if not isinstance(t, str):
                raise ValueError("Element type must be string")

            e = Gst.ElementFactory.make(t, name)

            # if t=='appsink':
            #     e.connect("new-sample", self.appsinkhandler, name)

            if e is None:
                raise ValueError("Nonexistant element type: " + t)
            self.weakrefs[str(e)] = e
            self.elementTypesById[id(e)] = t
            elementsByShortId[id(e)] = e

            for i in kwargs:
                v = kwargs[i]
                self.set_property(e, i, v)

            self.pipeline.add(e)
            op = []
            # May need to use an ID if its a remore command
            if connect_to_output:
                if not isinstance(connect_to_output, (list, tuple)):
                    cto = [connect_to_output]
                else:
                    cto = connect_to_output

                for connect_to_output in cto:
                    if connect_to_output is not False:
                        if isinstance(connect_to_output, int):
                            connect_to_output = elementsByShortId[connect_to_output]

                        if id(connect_to_output) not in self.elementTypesById:
                            raise ValueError(
                                "Cannot connect to the output of: "
                                + str(connect_to_output)
                                + ", no such element in pipeline."
                            )
                        op.append(connect_to_output)
            else:
                # One auto connect
                if connect_to_output is None:
                    op = [None]

            for connect_to_output in op:
                # Element doesn't have an input pad, we want this to be usable as a fake source to go after a real source if someone
                # wants to use it as a effect
                if t == "audiotestsrc":
                    connect_to_output = False

                # This could be the first element
                if self.elements and (connect_to_output is not False):
                    connect_to_output = connect_to_output or self.elements[-1]

                    # Fakesinks have no output, we automatically don't connect those
                    if self.elementTypesById[id(connect_to_output)] == "fakesink":
                        connect_to_output = False

                    # Decodebin doesn't have a pad yet for some awful reason
                    elif (
                        self.elementTypesById[id(connect_to_output)] == "decodebin"
                    ) or connect_when_available:
                        eid = time.time()
                        f = linkClosureMaker(
                            weakref.ref(self),
                            connect_to_output,
                            e,
                            connect_when_available,
                            eid,
                        )

                        self.waitingCallbacks[eid] = f
                        # Dummy 1 param because some have claimed to get segfaults without
                        connect_to_output.connect("pad-added", f, 1)
                    else:
                        link(connect_to_output, e)

            # Sidechain means don't set this element as the
            # automatic thing that the next entry links to
            if not sidechain:
                self.elements.append(e)
            else:
                self.sidechainElements.append(e)

            self.namedElements[name] = e

            self.lastElementType = t
            p = weakref.proxy(e)

            self.proxyToElement[id(p)] = e
            # List it under the proxy as well
            self.elementTypesById[id(p)] = t
            elementsByShortId[id(p)] = e

        # Mark as a JACK user so we can stop if needed for JACK
        # Stuff
        if t.startswith("jackaudio"):
            with lock:
                jackChannels[self.uuid] = weakref.ref(self)
        return p

    def add_elementRemote(self, *a, **k):
        return id(self.add_element(*a, **k))

    def add_jack_mixer_send_elements(self, target, idee, volume=-60):
        with self.lock:
            if not isinstance(target, str):
                raise ValueError("Target must be string")

            e = self.makeElement("tee")
            q = self.makeElement("queue")
            q2 = self.makeElement("queue")
            q2.max_size_buffers = 1
            q.max_size_buffers = 1
            q.leaky = 2
            q2.leaky = 2
            volume_element = self.makeElement("volume")
            volume_element.set_property("volume", 10 ** (volume / 20))

            e2 = self.makeElement("jackaudiosink", "_send" + str(len(self.elements)))
            e2.set_property("buffer-time", 10)
            e2.set_property("port-pattern", "fdgjkndgmkndfmfgkjkf")
            e2.set_property("sync", False)
            e2.set_property("slave-method", 0)
            e2.set_property("provide-clock", False)
            e2.set_property("connect", False)

            e2.latency_time = 10

            tee_src_pad_template = e.get_pad_template("src_%u")
            tee_audio_pad = e.request_pad(tee_src_pad_template, None, None)
            tee_audio_pad2 = e.request_pad(tee_src_pad_template, None, None)

            if self.elements:
                link(self.elements[-1], e)
            link(tee_audio_pad, q)
            link(tee_audio_pad2, q2)
            self.elements.append(q2)

            link(q, volume_element)
            link(volume_element, e2)

            return id(volume_element), id(e2)

    def set_property(self, element, prop, value):
        with self.lock:
            if isinstance(element, int):
                element = elementsByShortId[element]

            if prop == "location" and self.elementTypesById[id(element)] == "filesrc":
                if not os.path.isfile(value):
                    raise ValueError("No such file: " + value)

            if prop == "caps":
                value = Gst.Caps(value)
                self.weakrefs[str(value)] = value

            if prop.startswith("_"):
                prop = prop[1:]

            prop = prop.replace("_", "-")

            prop = prop.split(":")
            if len(prop) > 1:
                childIndex = int(prop[0])
                target = element.get_child_by_index(childIndex)
                target.set_property(prop[1], value)
                self.weakrefs[str(target) + "fromgetter"] = target
            else:
                element.set_property(prop[0], value)

    def isActive(self):
        with self.lock:
            if self.pipeline.get_state(1000_000_000)[1] == Gst.State.PAUSED:
                return True
            if self.pipeline.get_state(1000_000_000)[1] == Gst.State.PLAYING:
                return True


gstp = None

ppid = os.getppid()


def main():
    global gstp

    gstp = GStreamerPipeline()
    # Replace the dummy we put there for the linter
    rpc[0] = jsonrpyc.RPC(target=gstp)
    # def print(*a):
    #     rpc[0]("print", [str(a)])

    while 1:
        time.sleep(1)

        if (not check_pid(ppid)) or stopflag[0]:
            sys.exit()

        if not os.getppid() == ppid:
            return


if __name__ == "__main__":
    main()



import threading
import weakref
import traceback
import time
import re
import jack
import sys

_didPatch = False

# This is NixOS compatibility stuff, we could be running as an output from setup.py
# Or we could be running directly with python3 file.py
try:
    from . import jsonrpyc
except ImportError:
    import jsonrpyc


portInfoByID = weakref.WeakValueDictionary()


class PortInfo:
    def __init__(self, name, isInput, sname, isAudio, aliases=None):
        self.isOutput = self.is_output = not isInput
        self.isInput = self.is_input = isInput
        self.isAudio = self.is_audio = isAudio
        self.name = name
        self.shortname = sname
        self.clientName = name[: -len(":" + sname)]
        portInfoByID[id(self)] = self
        self.aliases = aliases or []

    def toDict(self):
        return {
            "name": self.name,
            "isInput": self.is_input,
            "sname": self.shortname,
            "isAudio": self.isAudio,
            "aliases": self.aliases,
        }


def portToInfo(p):
    return PortInfo(p.name, p.is_input, p.shortname, p.is_audio, list(p.aliases))


lock = threading.Lock()


def f():
    global _didPatch
    if True:
        if not _didPatch:

            def _get_ports_fix(
                self,
                name_pattern="",
                is_audio=False,
                is_midi=False,
                is_input=False,
                is_output=False,
                is_physical=False,
                can_monitor=False,
                is_terminal=False,
            ):
                if name_pattern:
                    re.compile(name_pattern)

                return jack.Client._get_ports(
                    self,
                    name_pattern,
                    is_audio,
                    is_midi,
                    is_input,
                    is_output,
                    is_physical,
                    can_monitor,
                    is_terminal,
                )

            jack.Client._get_ports = jack.Client.get_ports
            jack.Client.get_ports = _get_ports_fix
            _didPatch = True


f()

shouldExit = [0]

# We can't ship port objects on the wire, we have to do this instead.
# Anything that would normally take a port object must take a name instead.


class JackClientProxy:
    def __getattr__(self, attr):
        def f(*a, **k):
            x = getattr(self.clientObj, attr)(*a, **k)
            if isinstance(x, jack.Port):
                x = portToInfo(x).toDict()
            return x

        return f

    def exit(self, *a, **k):
        shouldExit[0] = 1
        sys.exit()

    def get_ports(self, *a, **k):
        x = self.clientObj.get_ports(*a, **k)
        x = [portToInfo(i).toDict() for i in x]
        return x

    def get_all_connections(self, p):
        p = self.clientObj.get_port_by_name(p)
        x = self.clientObj.get_all_connections(p)
        x = [portToInfo(i).toDict() for i in x]
        return x

    def init(self, *a, **k):
        self.clientObj = jack.Client(
            "Overseer" + str(time.monotonic()), no_start_server=True
        )
        self.clientObj.set_port_connect_callback(onPortConnect)
        self.clientObj.set_port_registration_callback(
            on_port_registered, only_available=False
        )
        self.clientObj.activate()

    def __init__(self) -> None:
        self.clientObj = None

    def disconnect(self, f, t):
        global realConnections
        if lock.acquire(timeout=30):
            try:
                f = self.clientObj.get_port_by_name(f)
                t = self.clientObj.get_port_by_name(t)
            except jack.JackError:
                return

            try:
                if not self.clientObj:
                    return
                try:
                    # This feels race conditionful but i think it is important so that we don't try to double-disconnect.
                    # Be defensive with jack, the whole thing seems britttle
                    self.clientObj.disconnect(f, t)

                except Exception:
                    pass
            finally:
                lock.release()

    def connect(self, f, t):
        global realConnections
        if lock.acquire(timeout=10):
            try:
                if not self.clientObj:
                    return

                # Ignore the nuisance of no longer existing ports. Airwires will get them if they come back.
                try:
                    f = self.clientObj.get_port_by_name(f)
                    t = self.clientObj.get_port_by_name(t)
                except jack.JackError:
                    return

                f_input = f.is_input

                if f.is_input:
                    if not t.is_output:
                        # Do a retry, there seems to be a bug somewhere
                        try:
                            f = self.clientObj.get_port_by_name(f.name)
                            t = self.clientObj.get_port_by_name(t.name)
                        except Exception:
                            return
                        if f.is_input:
                            if not t.is_output:
                                raise ValueError(
                                    "Cannot connect two inputs", str((f, t))
                                )
                else:
                    if t.is_output:
                        raise ValueError("Cannot connect two outputs", str((f, t)))
                f = f.name
                t = t.name
                try:
                    if f_input:
                        self.clientObj.connect(t, f)
                    else:
                        self.clientObj.connect(f, t)
                except Exception:
                    print(traceback.format_exc())
            finally:
                lock.release()


def on_port_registered(port, registered):
    if not port:
        return
    try:
        rpc.call(
            "on_port_registered",
            [port.name, port.is_input, port.shortname, port.is_audio, registered],
        )
    except Exception:
        print(traceback.format_exc())
        raise


def onPortConnect(a, b, c):
    rpc.call("on_port_connected", [a.is_output, a.name, b.name, c])


jackclient = None
rpc = None


def main():
    global jackclient
    global rpc
    jackclient = JackClientProxy()

    rpc = jsonrpyc.RPC(jackclient)

    import os
    import sys

    ppid = os.getppid()

    # https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python
    def check_pid(pid):
        """Check For the existence of a unix pid."""
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    while 1:
        time.sleep(10)
        if not check_pid(ppid):
            sys.exit()
        if not ppid == os.getppid():
            sys.exit()
        if shouldExit[0]:
            sys.exit()


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="icemedia",
    version="0.1.1",
    author="Daniel Dunn",
    author_email="dannydunn@eternityforest.com",
    description="A GStreamer wrapper with JACK connection management",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EternityForest/iceflow",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["JACK-Client", "scullery"],
)


# To push to pypi
# python3 setup.py sdist bdist_wheel
# python3 -m twine upload dist/*



import os
import time
import sys
import icemedia.sound_player

testmedia = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "Brothers Unite.opus"
)

icemedia.sound_player.play_sound(testmedia)
time.sleep(3)
sys.exit()



import gc
import unittest
import time
import icemedia.jack_tools
import icemedia.iceflow

"Only works if your device names match, may need to change it"

icemedia.jack_tools.start_managing()


class Player(icemedia.iceflow.GstreamerPipeline):
    def __init__(self):
        icemedia.iceflow.GstreamerPipeline.__init__(self, realtime=False)
        self.sink = self.add_element("audiotestsrc")
        self.sink = self.add_element("jackaudiosink", client_name="JackTest", connect=0)


class TestJackAudio(unittest.TestCase):
    "Note: Requires JACK or pipewire equivalent to be running"

    def test_airwire(self):
        p = Player()
        p.start()

        print("You should hear noise")
        aw = icemedia.jack_tools.Airwire("JackTest", "Built-in Audio Analog Stereo")
        aw.connect()
        time.sleep(1)
        print("No more noise")
        aw.disconnect()

        del aw
        gc.collect()
        time.sleep(1)
